/**
 * @author Aleksandr A. Lomov <lomov@cs.karelia.ru>
 * @date   24.05.2016
 * @brief  Interface for access point loader.
 *
 * This file is part of the CuteSIB project.
 *
 * @section LICENSE
 *
 * CuteSIB is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * CuteSIB is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CuteSIB; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301  USA
 *
 * @section DESCRIPTION
 *
 * Access point loader is a main plugin interface.
 * Loader is a factory to create access point instances.
 * Also loader contains information about access points (version, protocol etc.).
 *
 * The loader is used because Qt plug-in loader is not designed
 * to create several instances of plug-in.
 * Some details here:
 * http://stackoverflow.com/questions/17837383/how-to-create-multiple-instances-of-the-same-plugin-in-qt-using-qpluginloader
 *
 * Web-site: http://oss.fruct.org/wiki/SmartSpace-platform
 * Copyright (C) 2015-2016 by PetrSU.
 */


#ifndef CUTESIB_ACCESSPOINTLOADERINTERFACE_H
#define CUTESIB_ACCESSPOINTLOADERINTERFACE_H

#include "accesspointinterface_global.h"

#include <QtCore/QObject>
#include <QtCore/QString>

#include "accesspointinterface.h"


// Define indetifier for the access point interface.
#define AccessPointLoaderInterface_iid "org.fruct.oss.Projects.CuteSIB.AccessPointLoaderInterface"

namespace cutesib {
namespace access_point {

/**
 * @brief The AccessPointLoaderInterface class is used as a main interface of plugin (access point).
 * This iterface to implement a factory to create access point instances.
 */
class ACCESSPOINTINTERFACESHARED_EXPORT AccessPointLoaderInterface
{

public:
    /**
     * @brief Empty virtual destructor for an inheritance.
     */
    virtual ~AccessPointLoaderInterface() {}

    /**
     * @brief Gets identifier of loader.
     * Each loader and instance (access point) have own identifier.
     * Identifers are used by SIB to manage modules (plugins).
     * The identifier is defined by loader developer and must be uniq
     * for all loaders that are used by SIB.
     * @return loader identifier.
     */
    virtual QString id() = 0;

    /**
     * @brief Gets a description of access points that are created by loader.
     * @return description.
     */
    virtual QString description() = 0;

    /**
     * @brief Gets a protocols that are supported by access points.
     * Access point can support one or several protocols (application layer)
     * to recieve requests from clients.
     * For example, access point can support SSAP format: XML and JSON.
     * Access point dosen't implement details of protocols processing,
     * this is done by ProtocolHandlers on the SIB side.
     *
     * Format of this string is protocals names divided by commas: 'SSAP/XML,SSAP/JSON'.
     * @return rotocols supported by access points.
     */
    virtual QString protocols() = 0;

    /**
     * @brief Gets a version of loader and access points.
     * @return version.
     */
    virtual QString version() = 0;

    /**
     * @brief Creates an access point.
     * @param accessPointId is an identifier of new access point.
     * @param parameters for the access point.
     * @return created accsss point on success or nullptr otherwise.
     */
    virtual AccessPointInterface* create(QString accessPointId, QString parameters) = 0;

protected:
    /**
     * @brief Protected empty constructor to prevent instantiating.
     */
    AccessPointLoaderInterface() {}
};


} // namespace access_point
} // namespace cutesib

// Declare QT interface.
Q_DECLARE_INTERFACE(cutesib::access_point::AccessPointLoaderInterface, AccessPointLoaderInterface_iid)

#endif // CUTESIB_ACCESSPOINTLOADERINTERFACE_H
