/**
 * @author Aleksandr A. Lomov <lomov@cs.karelia.ru>
 * @date   02.03.2015
 * @brief  Base class for access point client.
 *
 * This file is part of the CuteSIB project.
 *
 * @section LICENSE
 *
 * CuteSIB is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * CuteSIB is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CuteSIB; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301  USA
 *
 * @section DESCRIPTION
 *
 * This class is used in the implementation of access point
 * to represent a client.
 *
 * Contact: http://oss.fruct.org/wiki/SmartSpace-platform
 * Copyright (C) 2015-2016 by PetrSU.
 */

#ifndef CUTESIB_ACCESSPOINTCLIENT_H
#define CUTESIB_ACCESSPOINTCLIENT_H

#include "accesspointinterface_global.h"
#include "accesspointinterface.h"

#include <QObject>

namespace cutesib {
namespace access_point {

// Forward declaration to compile (eliminate cross h. include)
class AccessPointInterface;

/**
 * @brief The base Client class to represent access point client.
 */
class ACCESSPOINTINTERFACESHARED_EXPORT Client : public QObject
{
    Q_OBJECT

public:

    /**
     * @brief Client class constructor.
     * @param parent for the instance.
     */
    explicit Client(QObject *parent = 0);

    /**
     * @brief Client class destructor.
     */
    virtual ~Client();

    /**
     * @brief Gets an access point of client.
     * @return access point.
     */
    virtual AccessPointInterface *getAccessPoint() = 0;

    /**
     * @brief Gets a node identifier of client.
     * When a clien joins to smart space a node identifier
     * from SSAP message is used for the client.
     * This method return this identifier.
     * @return node idetifier of the client.
     */
    QString nodeId();

    /**
     * @brief Sets a node identifier of client.
     * When a clien joins to smart space a node identifier
     * from SSAP message is used for the client.
     * This method sets this identifier.
     * @param nodeId is identifier of a client.
     */
    void setNodeId(QString nodeId);

    /**
     * @brief Sends a message to client using access point.
     * @param message Message to send.
     * @return 0 on success or not 0 otherwise.
     */
    virtual int send(QByteArray message) = 0;

    /**
     * @brief Checks client state: connected or not.
     * @return true when the client is connected or false otherwise.
     */
    virtual bool isConnected() = 0;

    /**
     * @brief Closes a connection with a client.
     * With closing connection a client will be removed from access point.
     * The client cannot be used for other action anymore.
     */
    virtual void close() = 0;


private: // Fields
    QString nodeId_;                        ///< Identifier of client (node id in Smart-M3).
    AccessPointInterface *accessPoint_;     ///< Access point of client.

    ///TODO: For debug, remove in stable.
#ifdef CUTESIB_DEBUG_INSTANCE_COUNTER
    static QAtomicInt instanceCounter;    // Counter of instances.
    static QAtomicInt maxInstanceCounter;    // Counter of instances.
#endif
};

} // namespace access_point
} // namespace cutesib

#endif // CUTESIB_ACCESSPOINTCLIENT_H
