/**
 * @author Aleksandr A. Lomov <lomov@cs.karelia.ru>
 * @date   13.03.2015
 * @brief  Class to represent request from client of access point to SIB.
 *
 * This file is part of the CuteSIB project.
 *
 * @section LICENSE
 *
 * CuteSIB is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * CuteSIB is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CuteSIB; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301  USA
 *
 * @section DESCRIPTION
 *
 * Class to represent request from client of access point to SIB.
 * Access point uses this class as a wrapper for received request
 * from client to forward it to SIB.
 *
 * Contact: http://oss.fruct.org/wiki/SmartSpace-platform
 * Copyright (C) 2015-2016 by PetrSU.
 */

#ifndef CUTESIB_REQUEST_H
#define CUTESIB_REQUEST_H


#include "accesspointinterface_global.h"
#include "accesspointinterface.h"


namespace cutesib {
namespace access_point {

// Forward declaration to compile (eliminate cross h. include)
class AccessPointInterface;
class Client;

class ACCESSPOINTINTERFACESHARED_EXPORT Request: public QObject
{
    Q_OBJECT

public:
    /**
     * @brief Request class constructor.
     * @param accessPoint that is received the request.
     * @param client who sent the request.
     * @param dataProtcol is protocol identifier (format) of request data (e.g. SSAP/XML).
     * @param data of request.
     * @param parent for the instance.
     */
    explicit Request(AccessPointInterface *accessPoint, QSharedPointer<Client> client, QString dataProtcol, QByteArray data, QObject *parent = 0);

    /**
     * @brief Request class destructor.
     */
    virtual ~Request();

    /**
     * @brief Gets access poin that is received the request.
     * @return access point.
     * @see AccessPointInterface
     */
    AccessPointInterface* accessPoint();

    /**
     * @brief Gets the client who sent the request.
     * @return client.
     * @see Client
     */
    QSharedPointer<Client> client();

    /**
     * @brief Gets protocol identifier (format) of request data (e.g. SSAP/XML).
     * @return protocol identifier.
     */
    QString dataProtocol() const;

    /**
     * @brief Gets request data.
     * @return request data.
     */
    const QByteArray data() const;


private:    // Fields.
    AccessPointInterface *accessPoint_; ///< Access poin that is received the request.
    QSharedPointer<Client> client_;     ///< Client who sent the request.
    QString dataProtocol_;              ///< Protocol idetifier (format) of the request data.
    QByteArray data_;                   ///< Data representing a request.

    ///TODO: For debug, remove in stable.
    static QAtomicInt instanceCounter;    // Counter of instances.
};

} // namespace access_point
} // namespace cutesib

#endif // CUTESIB_REQUEST_H
