/**
 * @author Aleksandr A. Lomov <lomov@cs.karelia.ru>
 * @date   25.09.2015
 * @brief  Request of command.
 *
 * This file is part of the CuteSIB project.
 *
 * @section LICENSE
 *
 * CuteSIB is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * CuteSIB is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CuteSIB; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301  USA
 *
 * @section DESCRIPTION
 *
 * CommandRequest is used when a module needs to delegate operation or send some notifications to other module.
 *
 * Contact: http://oss.fruct.org/wiki/SmartSpace-platform
 * Copyright (C) 2015-2016 by PetrSU.
 */


#include "commandrequest.h"


/*****************************************************************************/
/************************** Constructors/destructors *************************/

cutesib::CommandRequest::CommandRequest(CommandType type, QString from, QString to, bool needConfirmation, QObject *parent):
    Command(type, from, to, parent),
    needConfirmation_(needConfirmation),
    sharedInsertTriples_(&insertTriples_, deleteTriples),
    sharedRemoveTriples_(&removeTriples_, deleteTriples),
    sharedQueryTriples_(&queryTriples_, deleteTriples)
{ }


cutesib::CommandRequest::CommandRequest(cutesib::CommandType type, QString from, bool needConfirmation, QObject *parent) :
    Command(type, from, parent),
    needConfirmation_(needConfirmation),
    sharedInsertTriples_(&insertTriples_, deleteTriples),
    sharedRemoveTriples_(&removeTriples_, deleteTriples),
    sharedQueryTriples_(&queryTriples_, deleteTriples)
{ }


cutesib::CommandRequest::~CommandRequest()
{

}


/*****************************************************************************/
/****************************** Public methods *******************************/

bool cutesib::CommandRequest::needConfirmation() const
{
    return needConfirmation_;
}


void cutesib::CommandRequest::setConfirmation(bool needConfirmation)
{
    needConfirmation_ = needConfirmation;
}


void cutesib::CommandRequest::setInsertTriples(QList<cutesib::Triple *> triples)
{
#if QT_VERSION >= 0x050000
    sharedInsertTriples_.reset();
#else
    sharedInsertTriples_.clear();
#endif
    insertTriples_ = triples;
    sharedInsertTriples_ = QSharedPointer<QList<Triple *>>(&insertTriples_, deleteTriples);
}

void cutesib::CommandRequest::setRemoveTriples(QList<cutesib::Triple *> triples)
{
#if QT_VERSION >= 0x050000
    sharedRemoveTriples_.reset();
#else
    sharedRemoveTriples_.clear();
#endif
    removeTriples_ = triples;
    sharedRemoveTriples_ = QSharedPointer<QList<Triple *>>(&removeTriples_, deleteTriples);
}

void cutesib::CommandRequest::setQueryTriples(QList<cutesib::Triple *> triples)
{
#if QT_VERSION >= 0x050000
    sharedQueryTriples_.reset();
#else
    sharedQueryTriples_.clear();
#endif
    queryTriples_ = triples;
    sharedQueryTriples_ = QSharedPointer<QList<Triple *>>(&queryTriples_, deleteTriples);
}


void cutesib::CommandRequest::setSubscriptionId(QString id)
{
    subscriptionId_ = id;
}


const QList<cutesib::Triple *> cutesib::CommandRequest::insertTriples() const
{
    return *sharedInsertTriples_.data();

}


const QList<cutesib::Triple *> cutesib::CommandRequest::removeTriples() const
{
    return *sharedRemoveTriples_.data();
}


const QList<cutesib::Triple *> cutesib::CommandRequest::queryTriples() const
{
    return *sharedQueryTriples_.data();
}


void cutesib::CommandRequest::setInsertTriplesS(QSharedPointer<QList<cutesib::Triple *> > triples)
{
    sharedInsertTriples_ = triples;
}


void cutesib::CommandRequest::setRemoveTriplesS(QSharedPointer<QList<cutesib::Triple *> > triples)
{
    sharedRemoveTriples_ = triples;
}


void cutesib::CommandRequest::setQueryTriplesS(QSharedPointer<QList<cutesib::Triple *> > triples)
{
    sharedQueryTriples_ = triples;
}


const QString cutesib::CommandRequest::subscriptionId() const
{
    return subscriptionId_;
}


/*****************************************************************************/
/***************************** Private methods *******************************/

void cutesib::CommandRequest::deleteTriples(QList<cutesib::Triple *> *triples)
{
    foreach (Triple *t, *triples) {
        delete t;
    }

    triples->clear();
}
