/**
 * @author Aleksandr A. Lomov <lomov@cs.karelia.ru>
 * @date   08.03.2015
 * @brief  Main file to start CuteSIB.
 *
 * This file is part of the CuteSIB project.
 *
 * @section LICENSE
 *
 * CuteSIB is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * CuteSIB is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CuteSIB; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301  USA
 *
 * @section DESCRIPTION
 *
 * Makes some configuration actions and starts the SIB.
 *
 * Options:
 * -h, --help     Displays help.
 * -v, --version  Displays version information.
 *
 * Arguments:
 * config         Configuration file to use (path to file with filename).
 *
 * Web-site: http://oss.fruct.org/wiki/SmartSpace-platform
 * Copyright (C) 2016 by PetrSU.
 */


#include <iostream>
#include <signal.h>

#if defined(WINDOWS) || defined(WIN32)
#include "windows.h"
#endif

#include <QtCore/QtGlobal>
#include <QtCore/QCoreApplication>
#include <QtCore/QDir>
#include <QtCore/QString>
#include <QtCore/QStringList>
#include <QtCore/QtPlugin>
#include <QtCore/QPluginLoader>

#if QT_VERSION >= 0x050200
#include <QtCore/QCommandLineParser>
#endif

#include "siberrors.h"
#include "sibmanager.h"
#include "accesspointloaderinterface.h"
#include "logger.h"


// To import static plgins.
#ifdef CUTESIB_TCPAP_STATIC
Q_IMPORT_PLUGIN(TcpAccessPointLoader)
#endif
#ifdef CUTESIB_DISCOVERYAP_STATIC
Q_IMPORT_PLUGIN(DiscoveryAccessPointLoader)
#endif
#ifdef CUTESIB_HTTPAP_STATIC
Q_IMPORT_PLUGIN(HttpAccessPointLoader)
#endif
#ifdef CUTESIB_TCPASYNCAP_STATIC
Q_IMPORT_PLUGIN(TcpAsyncAccessPointLoader)
#endif

///TODO: Where type should be registered?
// To gegister a metatype for signal-slot queue connection.
Q_DECLARE_METATYPE(QSharedPointer<cutesib::access_point::Request>)
Q_DECLARE_METATYPE(QSharedPointer<cutesib::access_point::Client>)

// Constants for help information.
#if defined(WINDOWS) || defined(WIN32) || defined(WIN64)
#define CUTESIB_USAGE "Usage: cutesib.exe [options] [config]\n"
#else
#define CUTESIB_USAGE "Usage: ./cutesib [options] [config]\n"
#endif

#define CUTESIB_DESCRIPTION  CUTESIB_NAME" - Semantic Information Brocker for Smart-M3 platform.\n\n" \
    "Supported protocols:\n" \
    " - SSAP/XML.\n\n" \
    "Supported operations:\n" \
    " - join - now does nothing;\n" \
    " - leave - connection will be ended;\n" \
    " - insert/remove/update/query triples;\n" \
    " - subscribe/unsubscribe triples;\n" \
    " - SPARQL queries (SELECT, ASK, CONSTRUCT).\n\n" \
    "Built-in plugins (access points):\n"

#define CUTESIB_OPTIONS "\n\nOptions:\n" \
    "  -h, --help     Displays this help.\n" \
    "  -v, --version  Displays version information.\n\n" \
    "Arguments:\n" \
    "   config         Configuration file to use (path to file with filename).\n"


/**
 * @brief Quits the application.
 * @param signal identifier for quit the application.
 */
static void quit(int signal);

/**
 * @brief Gets information about static plugins (built-in access points).
 * @return string with descriptions.
 */
static QString getPluginsDescriptions();

// Parsing parametrs with QCommandLineParser or manually for old Qt versions.
#if QT_VERSION >= 0x050200
static void processParameters(QCoreApplication &app, QString description, QString &configPath);
#else
static void processParameters(QString description, QString &configPath);
#endif


// Signals for Windows OS.
#if defined(WINDOWS) || defined(WIN32)
/**
 * @brief Handler for Windows signals.
 * @param fdwCtrlType is a type if event.
 * @return true if the function handles the control signal, otherwise false.
 */
static BOOL signalHandler(DWORD fdwCtrlType);

/**
 * @brief Registers a signal handler.
 */
static void registerSignalHandler(void);

#else // Signals for Unix/Linux.

/**
 * @brief Registers a signal handler (Unix/Linux).
 * @param quitSignals are signals to quit the application.
 * @param ignoreSignals are signals ignored by handler.
 */
static void registerSignalHandler(const std::vector<int>& quitSignals,
                             const std::vector<int>& ignoreSignals = std::vector<int>());

#endif



int main(int argc, char *argv[])
{
    using cutesib::Logger;
    using cutesib::SibManager;

    QCoreApplication app(argc, argv);
    QCoreApplication::setApplicationName(CUTESIB_NAME);
    QCoreApplication::setApplicationVersion(CUTESIB_VERSION);

    // Create description for help.
    QString description(CUTESIB_DESCRIPTION);
    QString pluginDescription = getPluginsDescriptions();

    if (pluginDescription.isEmpty()) {
        description += " - No built-in plugins.";
    } else {
        description += pluginDescription;
    }

    QString configPath;

    // Process parameters (show help, version, get config path).
#if QT_VERSION >= 0x050200
    processParameters(app, description, configPath);
#else
    processParameters(description, configPath);
#endif

    // Register handlers for signals (CTRL+C).
#if defined(WINDOWS) || defined(WIN32)
    registerSignalHandler();
#else
    registerSignalHandler({SIGQUIT, SIGINT, SIGTERM, SIGHUP});
#endif

    // Sets handler for debug messages.
#if QT_VERSION >= 0x050000
    qInstallMessageHandler(Logger::messageHandler);
#else
    qInstallMsgHandler(Logger::messageHandler);
#endif

    Logger::message() << "Starting: "<< CUTESIB_NAME << " version " << CUTESIB_VERSION << endl;

    SibManager sibmanager(configPath);

    QObject::connect(&app, SIGNAL(aboutToQuit()), &sibmanager, SLOT(shutdown()));

    cutesib::SibError result = sibmanager.initialize();

    if (result != cutesib::SibError::NONE) {
#if __GNUC__ < 5 && __GNUC_MINOR__ < 7
        qCritical() << CUTESIB_NAME " cannot be started. Errno = "
                    << static_cast<int>(result);
#else
        qCritical() << CUTESIB_NAME " cannot be started. Errno = "
                    << static_cast<std::underlying_type<cutesib::SibError>::type>(result);
#endif
        return 0;
    }

    ///TODO: Where type should be registered?
    qRegisterMetaType<QSharedPointer<cutesib::access_point::Request>>();
    qRegisterMetaType<QSharedPointer<cutesib::access_point::Client>>();

    sibmanager.start();

    return app.exec();
}


static void quit(int signal)
{
    qDebug("\nQuit the application (user request signal = %d).\n", signal);

    cutesib::Logger::message() << "\nHave a nice day and say 'Goodbye to " << CUTESIB_NAME << "!'" << endl;
    QCoreApplication::quit();
}


static QString getPluginsDescriptions() {

    QString description;

    foreach (QObject *plugin, QPluginLoader::staticInstances()) {
        if (plugin) {
            auto *loader = qobject_cast<cutesib::access_point::AccessPointLoaderInterface*>(plugin);

            if (!description.isEmpty()) {
                description += ";\n";
            }
            description += " - " + loader->id() + " v" + loader->version() + ": "
                    + loader->description() + " (" + loader->protocols() + ")";
        }
    }

    return description + ".";
}


// Parsing parametrs with QCommandLineParser or manually for old Qt versions.
#if QT_VERSION >= 0x050200
static void processParameters(QCoreApplication &app, QString description, QString &configPath)
{
    QCommandLineParser parser;

    parser.setApplicationDescription(description);

    parser.addHelpOption();
    parser.addVersionOption();

    parser.addPositionalArgument("config",
                                 "Configuration file to use (path to file with filename).",
                                 "[config]");

    parser.process(app);

    // Check parameter for config file.
    const QStringList positionalArguments = parser.positionalArguments();

    if (!positionalArguments.isEmpty()) {
        configPath = positionalArguments.first();
    }
}
#else
static void processParameters(QString description, QString &configPath)
{
    description += CUTESIB_OPTIONS;

    QStringList args = QCoreApplication::arguments();

    args.removeAt(0); // Remove the programm name.

    // Check options (version, help).
    foreach(QString argument, args) {
        if (argument == "-h" || argument == "--help") {
            std::cout << CUTESIB_USAGE << description.toStdString();
            exit(EXIT_SUCCESS);
        }

        if (argument == "-v" || argument == "--version") {
            std::cout << CUTESIB_NAME << " " << CUTESIB_VERSION;
            QCoreApplication::exit(0);
            exit(EXIT_SUCCESS);
        }
    }

    // Check arguments (config path) after options.
    foreach(QString argument, args) {
        if (argument != "-h" && argument != "--help"
                && argument != "-v" && argument != "--version") {
            configPath = argument;
            return;
        }
    }
}
#endif


// Signals for Windows OS.
#if defined(WINDOWS) || defined(WIN32)

BOOL signalHandler(DWORD fdwCtrlType){

    switch( fdwCtrlType ){
    case CTRL_C_EVENT:
        quit(CTRL_C_EVENT);
        return( TRUE );
    case CTRL_CLOSE_EVENT:
        quit(CTRL_CLOSE_EVENT);
        return( TRUE );
    case CTRL_BREAK_EVENT:
        return FALSE;
    case CTRL_LOGOFF_EVENT:
        return FALSE;
    case CTRL_SHUTDOWN_EVENT:
        return FALSE;
    default:
        return FALSE;
    }
}


static void registerSignalHandler(void) {
    if( !SetConsoleCtrlHandler( (PHANDLER_ROUTINE) signalHandler, TRUE ) )
        printf("Unable to register Control Handler.\n");
}

#else // Signals for Unix/Linux.

static void registerSignalHandler(const std::vector<int>& quitSignals,
                             const std::vector<int>& ignoreSignals) {

    // All these signals will be ignored.
    for ( int sig : ignoreSignals )
        signal(sig, SIG_IGN);

    // Each of these signals calls the handler (quits the QCoreApplication).
    for ( int sig : quitSignals )
        signal(sig, quit);
}

#endif
