/**
 * @author Aleksandr A. Lomov <lomov@cs.karelia.ru>
 * @date   01.04.2015
 * @brief  Scheduler manages commands processing.
 *
 * This file is part of the CuteSIB project.
 *
 * @section LICENSE
 *
 * CuteSIB is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * CuteSIB is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CuteSIB; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301  USA
 *
 * @section DESCRIPTION
 *
 * Scheduler manages commands in the SIB. Modules dosen't send commands
 * directly beetwen each other. Commands are sent to scheduler and it
 * gets information from the command to resend the command to appropriate
 * reciever (other module).
 *
 * Contact: http://oss.fruct.org/wiki/SmartSpace-platform
 * Copyright (C) 2015-2016 by PetrSU.
 */


#ifndef CUTESIB_SCHEDULER_H
#define CUTESIB_SCHEDULER_H

#include <QObject>

#include "command.h"
#include "commandresponse.h"
#include "commandrequest.h"
#include "context.h"


namespace cutesib {

// Forward declaration.
class CommandResponse;


/**
 * @brief The Scheduler class  manages commands in the SIB. Modules dosen't send commands
 * directly beetwen each other. Commands are sent to scheduler and it
 * gets information from the command to resend the command to appropriate
 * reciever (other module).
 */
class Scheduler : public QObject
{
    Q_OBJECT

public:
    /**
     * @brief Scheduler class constructor.
     * @param context of the SIB.
     * @param parent of the instance.
     */
    Scheduler(Context *context, QObject *parent = 0);

    /**
     * @brief Scheduler class destructor.
     */
    ~Scheduler();

    /**
     * @brief Porform command by forward it to apropriate module.
     * If the command request does not contain information about destination module,
     * then mapping between command types and modules are used to forward request.
     * If the information exists, then request is forwarded to this module.
     * @param command request to perform.
     * @return response for command request or nullptr if the command does not need response.
     * @see Command
     * @see CommandRequest
     * @see CommandResponse
     */
    CommandResponse* perform(CommandRequest *command);


private: // Methods
    /**
     * @brief Hadles unmapped command.
     * Method is coalled when the destination module for request does not set
     * and mapping does not exists.
     * Request will be deleted if it does not require response.
     * @param command request to handle.
     * @return error response or nullptr when response does not require.
     */
    CommandResponse* handleUnmapped(CommandRequest *command);


private: // Fieldss.
    Context *context_;  ///< SIB context.

};

}

#endif // CUTESIB_SCHEDULER_H
