/**
 * @author Aleksandr A. Lomov <lomov@cs.karelia.ru>
 * @date   08.04.2015
 * @brief  Provides access to config file.
 *
 * This file is part of the CuteSIB project.
 *
 * @section LICENSE
 *
 * CuteSIB is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * CuteSIB is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CuteSIB; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301  USA
 *
 * @section DESCRIPTION
 *
 * Provides access to config file.
 * Config file contains information about access poits (build-in/static or dynamic)
 * and about triplestore.
 *
 * Contact: http://oss.fruct.org/wiki/SmartSpace-platform
 * Copyright (C) 2015-2016 by PetrSU.
 */


#include "sibconfig.h"

#include <QDebug>


/*****************************************************************************/
/*************************** Definitions/Constants ***************************/

// Section and section prefixes for ini-file.
const QString cutesib::SibConfig::SectionTriplestor = "Triplestore";
const QString cutesib::SibConfig::SectionPrefixAccessPoint = "AccessPoint:";

// Keys/properties for ini-file.
const QString cutesib::SibConfig::KeyName = "name";
const QString cutesib::SibConfig::KeyFullpath = "fullpath";
const QString cutesib::SibConfig::KeyAasyncFlag = "async";
const QString cutesib::SibConfig::KeyPath = "path";
const QString cutesib::SibConfig::KeyId = "id";
const QString cutesib::SibConfig::KeyParameters = "parameters";
const QChar cutesib::SibConfig::ParametersDelimeter = ',';

const QString cutesib::SibConfig::DefaultFolder = "./";


/*****************************************************************************/
/************************** Constructors/destructors *************************/

cutesib::SibConfig::SibConfig(QString path) : settings_(path, QSettings::IniFormat)
{
    foreach (QString groupName, settings_.childGroups()) {
        if (groupName.startsWith(SectionPrefixAccessPoint)) {
            accessPoints_.append(groupName.remove(SectionPrefixAccessPoint));
        }
    }

    qDebug() << "Configured access points: " << accessPoints_;
}

cutesib::SibConfig::~SibConfig()
{

}


/*****************************************************************************/
/****************************** Public methods *******************************/

QStringList cutesib::SibConfig::getAccessPointsConfigIds() const
{
    return accessPoints_;
}


QString cutesib::SibConfig::getAccessPointParameters(QString configId)
{
    settings_.beginGroup(SectionPrefixAccessPoint + configId);

    QVariant parametersList = settings_.value(KeyParameters);
    QString parameters;
    if (parametersList.type() == QVariant::StringList) {
        parameters = parametersList.toStringList().join(ParametersDelimeter);
    } else {
        parameters = parametersList.toString();
    }

    settings_.endGroup();

    return parameters;
}


QString cutesib::SibConfig::getAccessPointPath(QString configId)
{
    settings_.beginGroup(SectionPrefixAccessPoint + configId);
    QString path = settings_.value(KeyFullpath).toString();

    if (path.isEmpty()) {
        path = settings_.value(KeyPath).toString() + settings_.value(KeyName).toString() + CUTESIB_DLL_EXT;
    }

    settings_.endGroup();

    return path.isEmpty() ? DefaultFolder :  path;
}


bool cutesib::SibConfig::getAccessPointAsyncFlag(QString configId)
{
    settings_.beginGroup(SectionPrefixAccessPoint + configId);
    int value = settings_.value(KeyAasyncFlag).toInt();
    settings_.endGroup();

    return value == 0 ? false : true;
}


QString cutesib::SibConfig::getAccessPointStaticId(QString configId)
{
    settings_.beginGroup(SectionPrefixAccessPoint + configId);

    QString id = settings_.value(KeyId).toString();

    settings_.endGroup();

    return id;
}


QString cutesib::SibConfig::getTriplestoreParameter(QString name)
{
    settings_.beginGroup(SectionTriplestor);

    QString value = settings_.value(name).toString();

    settings_.endGroup();

    ///TODO: Tripletore type (remove when other types will be supported.)
    // On Windows virtuoso is not supported.
#if defined(WINDOWS) || defined(WIN32) || defined(WIN64)
    QString type("virtuoso");

    if (type.compare(value, Qt::CaseInsensitive) == 0) {
        value = "hash-bdb";
    }
#endif

    return value;
}


QString cutesib::SibConfig::getModuleId(cutesib::CommandType type)
{
    return commandsMapping_[type];
}


void cutesib::SibConfig::setCommandMapping(QMap<cutesib::CommandType, QString> commandsMapping)
{
    commandsMapping_ = commandsMapping;
}
