/**
 * @author Aleksandr A. Lomov <lomov@cs.karelia.ru>
 * @date   08.04.2015
 * @brief  Provides access to config file.
 *
 * This file is part of the CuteSIB project.
 *
 * @section LICENSE
 *
 * CuteSIB is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * CuteSIB is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CuteSIB; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301  USA
 *
 * @section DESCRIPTION
 *
 * Provides access to config file.
 * Config file contains information about access poits (build-in/static or dynamic)
 * and about triplestore.
 *
 * Contact: http://oss.fruct.org/wiki/SmartSpace-platform
 * Copyright (C) 2015-2016 by PetrSU.
 */


#ifndef CUTSIB_SIBCONFIG_H
#define CUTSIB_SIBCONFIG_H

#include <QList>
#include <QMap>
#include <QString>
#include <QStringList>
#include <QSettings>

#include "command.h"


namespace cutesib  {

/**
 * @brief The SibConfig class reads config file
 * and provides access to configuration parameters.
 * Now config file contains information about access poins and triple store.
 */
class SibConfig
{
public:
    /**
     * @brief SibConfig class constructor.
     * Read config file to get onformation about configuration of the SIB.
     * If the given file not found then it uses default values.
     * @param path to config file.
     */
    SibConfig(QString path);

    /**
     * @brief SibConfig class destructor.
     */
    ~SibConfig();

    /**
     * @brief Gets access points (ids) configured in config file.
     * In config file section for access points contains prefix and id ([prefix:configId]).
     * Method reads section for access points and gets ids.
     * @return config ids of all access points.
     */
    QStringList getAccessPointsConfigIds() const;

    /**
     * @brief Gets parameters of access point.
     * @param configId of access point in config file.
     * @return parameters for access point.
     */
    QString getAccessPointParameters(QString configId);

    /**
     * @brief Gets path to access point plug-in file (dynamic library).
     * @param configId of access point in config file.
     * @return path to file.
     */
    QString getAccessPointPath(QString configId);

    /**
     * @brief Gets async flag (async) of access point.
     * If the flag is equals 1, then the AP should be
     * moved to other (not main SIB) thread.
     * @param configId of access point in config file.
     * @return asynchronous flag value (async).
     */
    bool getAccessPointAsyncFlag(QString configId);

    /**
     * @brief Gets static id of access point.
     * Static id means id of build-in access point (static plugin).
     * @param configId of access point in config file.
     * @return id of build-in access point.
     */
    QString getAccessPointStaticId(QString configId);

    /**
     * @brief Gets a triplestore paramater value.
     * Founds paramater in the 'Triplestore' section of config file and returns a value of paramater.
     * If the parameter cannot be found the method return 'null'.
     * @param name a parameter name to get value.
     * @return parameter value.
     */
    QString getTriplestoreParameter(QString name);

    /**
     * @brief Gets identifier of module mapped to command type.
     * Modules performs some commands.
     * This method founds a module that is mapped to perform given command type.
     * @param type of command.
     * @return module id that performs given type of command.
     */
    QString getModuleId(CommandType type);

    /**
     * @brief Sets mapping between command types and modules (modules ids).
     * Now mapping is done in #SibManager not in config file.
     * @param commandsMapping is mapping between command types and modules.
     */
    void setCommandMapping(QMap<CommandType, QString> commandsMapping);

private: // Fields.
    // Sections of config file.
    static const QString SectionTriplestor;             ///< Triple store section name.
    static const QString SectionPrefixAccessPoint;      ///< Prefix for access point section ([prefix:id]).

    // Keys/Properties of config file.
    static const QString KeyName;       ///< Name key.
    static const QString KeyFullpath;   ///< Fullpath key.
    static const QString KeyAasyncFlag; ///< Async key (async=).
    static const QString KeyPath;       ///< Path key.
    static const QString KeyId;         ///< Id key.
    static const QString KeyParameters; ///< Parameters key.
    static const QChar ParametersDelimeter; ///< Delimeter for parameters key.


    static const QString DefaultFolder; ///< Default folder - './'.

    QSettings settings_;        ///< To access ini file.
    QStringList accessPoints_;  ///< Config ids of access points.
    QMap<CommandType, QString> commandsMapping_;    ///< mapping between command types and modules (modules ids).
};

} // Namespace cutesib.

#endif // CUTSIB_SIBCONFIG_H
