/**
 * @author Aleksandr A. Lomov <lomov@cs.karelia.ru>
 * @date   09.05.2015
 * @brief  Class to manage the SIB.
 *
 * This file is part of the CuteSIB project.
 *
 * @section LICENSE
 *
 * CuteSIB is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * CuteSIB is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CuteSIB; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301  USA
 *
 * @section DESCRIPTION
 *
 * Class to manage SIB.
 * It performs all action to initialize the SIB and prepares SIB for working or shutdown.
 *
 * Contact: http://oss.fruct.org/wiki/SmartSpace-platform
 * Copyright (C) 2015-2016 by PetrSU.
 */


#include "sibmanager.h"

#include <QtCore/QDir>
#include <QtCore/QDebug>

#include "sibconfig.h"
#include "siberrors.h"
#include "logger.h"
#include "context.h"
#include "triplestore.h"
#include "basicoperationhandler.h"
#include "sparqlhandler.h"


/*****************************************************************************/
/************************** Constructors/destructors *************************/

cutesib::SibManager::SibManager(QObject *parent) : QObject(parent)
{
    config_ = nullptr;
    accessPointManager_ = nullptr;
    scheduler_ = nullptr;

    configPath_ = QDir::currentPath() + QDir::separator() + CUTESIB_CONFIG_NAME;
}


cutesib::SibManager::SibManager(QString configPath, QObject *parent) : QObject(parent)
{
    config_ = nullptr;
    accessPointManager_ = nullptr;
    scheduler_ = nullptr;

    configPath_ = configPath;
}


cutesib::SibManager::~SibManager()
{
    delete accessPointManager_;
    delete scheduler_;
    delete triplestore_;
    delete config_;
}


/*****************************************************************************/
/****************************** Public methods *******************************/

cutesib::SibError cutesib::SibManager::initialize()
{
    QString configFoundPath = findConfig();

    if (configFoundPath == QString::null) {
        return SibError::NOT_FOUND_CONFIG;
    }
    
    config_ = new SibConfig(configFoundPath);
    context_.setConfig(config_);

    scheduler_ = new Scheduler(&context_, this);
    context_.setScheduler(scheduler_);

    triplestore_ = new Triplestore(config_);
    triplestore_->initialize();
    context_.setTriplestore(triplestore_);

    accessPointManager_ = new AccessPointManager(config_, this);
    accessPointManager_->load();

    prepareProtocolsHandlers();
    prepareOperationsHandlers();

    config_->setCommandMapping(operationMapping_);

    QList<SibModule*> modules;

    foreach (SibModule *module, operationHandlers_.values()) {
        modules.append(module);
    }

    foreach (SibModule *module, protocolHandlers_.values()) {
        modules.append(module);
    }

    context_.setModules(modules);

    foreach (SibModule *module, modules) {
        SibError result = module->initialize(&context_);

        if (result != SibError::NONE) {
            shutdown();
            qCritical() << "Module " << module->id() << "canot be initialized.";
            return result;
        }
    }

    foreach (ProtocolHandler *handler, protocolHandlers_) {
        accessPointManager_->setAccessPoints(handler);
    }

    connect(basicOperationHandler_, SIGNAL(queryResult(CommandResponse*)), ssapHandler_, SLOT(handleParallelQueryResponse(CommandResponse*)));

    return SibError::NONE;
}


cutesib::SibError cutesib::SibManager::start()
{

    foreach (ProtocolHandler *module, protocolHandlers_.values()) {
        if (!module->start()) {
            qWarning() << "Starting" << module->id() << " module failed.";

            return SibError::FALED;
        }
    }


    return SibError::NONE;
}


void cutesib::SibManager::stop()
{
    foreach (SibModule *module, context_.modules()) {
        module->stop();
    }
}


/*****************************************************************************/
/******************************* Public slots ********************************/

void cutesib::SibManager::shutdown()
{
    qDebug() << "Shutdown.";

    stop();

    foreach (SibModule *module, context_.modules()) {
        module->shutdown();
        delete module;
    }

    accessPointManager_->unloadAll();
}


/*****************************************************************************/
/***************************** Private methods *******************************/

void cutesib::SibManager::prepareProtocolsHandlers()
{
    ProtocolHandler *ssapHandler = new SsapHandler(config_, this);
    ssapHandler_ = (SsapHandler*) ssapHandler;
    protocolHandlers_.insert(ssapHandler->id(), ssapHandler);
}


void cutesib::SibManager::prepareOperationsHandlers()
{
    BasicOperationHandler *basicOperationHandler = new BasicOperationHandler(this);
    basicOperationHandler_ = basicOperationHandler;
    operationHandlers_.insert(basicOperationHandler->id(), basicOperationHandler);

    SubscriptionHandler *subscriptionHandler = new SubscriptionHandler(this);
    operationHandlers_.insert(subscriptionHandler->id(), subscriptionHandler);

    SparqlHandler* sparqlHandler = new SparqlHandler(this);
    operationHandlers_.insert(sparqlHandler->id(), sparqlHandler);

    // Connect signal aboud store updates.
    foreach (OperationHandler *op, operationHandlers_.values()) {
        if (op->needUpdateNotifications()) {
            foreach (OperationHandler *opSignal, operationHandlers_.values()) {
                connect(opSignal, SIGNAL(storeUpdated(CommandRequest*)),
                        op, SLOT(updateState(CommandRequest*)));
            }
        }
    }

    // Set command types mapping.
    operationMapping_.insert(CommandType::INSERT, basicOperationHandler->id());
    operationMapping_.insert(CommandType::REMOVE, basicOperationHandler->id());
    operationMapping_.insert(CommandType::UPDATE, basicOperationHandler->id());
    operationMapping_.insert(CommandType::QUERY, basicOperationHandler->id());

    operationMapping_.insert(CommandType::SUBSCRIBE, subscriptionHandler->id());
    operationMapping_.insert(CommandType::UNSUBSCRIBE, subscriptionHandler->id());

    operationMapping_.insert(CommandType::SPARQL_QUERY, sparqlHandler->id());
}


QString cutesib::SibManager::findConfig()
{
    QFileInfo checkPath(configPath_);

    if (checkPath.exists() && checkPath.isFile()) {
        Logger::message() << "Config file is found: " << checkPath.absoluteFilePath() << endl;
        return checkPath.absoluteFilePath();
    }

    QFileInfo checkCurrent(QDir::currentPath() + QDir::separator() + CUTESIB_CONFIG_NAME);

    if (checkCurrent.exists() && checkCurrent.isFile()) {
        Logger::message() << "Config file is found in the current directory: " << checkCurrent.absoluteFilePath() << endl;
        return checkCurrent.absoluteFilePath();
    }

    QFileInfo checkHome(QDir::homePath() + QDir::separator() + ".cutesib" + QDir::separator()  + CUTESIB_CONFIG_NAME);

    if (checkHome.exists() && checkHome.isFile()) {
        Logger::message() << "Config file is found in the home directory" << checkHome.absoluteFilePath() << endl;
        return checkHome.absoluteFilePath();
    }

#ifdef CUTESIB_CONFIG_PATH
    QString predefinePath(CUTESIB_CONFIG_PATH);
    QFileInfo predefineCheck(predefinePath);

    // Check if file exists and is not a directory?
    if (predefineCheck.exists() && predefineCheck.isFile()) {
        Logger::message() << "Config file is found: " << predefinePath << endl;
        return predefinePath;
    }
#endif

    Logger::message() << "Config file not found."<< endl;

    return QString::null;
}
