/**
 * @author Aleksandr A. Lomov <lomov@cs.karelia.ru>
 * @date   09.05.2015
 * @brief  Class to manage the SIB.
 *
 * This file is part of the CuteSIB project.
 *
 * @section LICENSE
 *
 * CuteSIB is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * CuteSIB is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CuteSIB; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301  USA
 *
 * @section DESCRIPTION
 *
 * Class to manage SIB.
 * It performs all action to initialize the SIB and prepares SIB for working or shutdown.
 *
 * Contact: http://oss.fruct.org/wiki/SmartSpace-platform
 * Copyright (C) 2015-2016 by PetrSU.
 */


#ifndef CUTESIB_SIBMANAGER_H
#define CUTESIB_SIBMANAGER_H

#include <QtCore/QObject>
#include <QtCore/QFileInfo>
#include <QtCore/QHash>

#include "accesspointmanager.h"
#include "context.h"
#include "sibconfig.h"
#include "scheduler.h"
#include "triplestore.h"
#include "ssaphandler.h"
#include "basicoperationhandler.h"
#include "subscriptionhandler.h"
#include "protocolhandler.h"

#define CUTESIB_CONFIG_NAME "cutesib.ini"


namespace cutesib {

/**
 * @brief The SibManager class performs all action to initialize the SIB
 * and prepares SIB for working or shutdown.
 * To configure the SIB the config file is used.
 * SibManager tries to load config file using paths in such order:
 * 1) given in constructor;
 * 2) current directory (where CuteSIB executable file is located);
 * 3) user home path (in '.cutesib' directory);
 * 4) defined with CUTESIB_CONFIG_PATH in compile-time (e.g. in .pro file).
 */
class SibManager : public QObject
{
    Q_OBJECT

public:
    /**
     * @brief SibManager class constructor.
     * @param parent of the instance.
     */
    explicit SibManager(QObject *parent = 0);

    /**
     * @brief SibManager class constructor with config file.
     * Given config file path is used to load config.
     * This path has a higher priority search the config file.
     * @param configPath is path to config file with filename.
     * @param parent of the instance.
     */
    explicit SibManager(QString configPath, QObject *parent = 0);

    /**
     * @brief SibManager class destructor.
     */
    ~SibManager();

    /**
     * @brief Initializes the SIB to start.
     * Makes SIB configuration, prepares triple-store, access points and handlers.
     * @return SibError::NONE on success or error code otherwise.
     * @see cutesib::SibError
     */
    SibError initialize();

    /**
     * @brief Starts SIB.
     * Starts all modules of the SIB and access points.
     * Call this method after #initialize().
     * @return SibError::NONE on success or SibError::FALED otherwise.
     * @see initialize()
     * @see stop()
     * @see cutesib::SibError
     */
    SibError start();

    /**
     * @brief Stops SIB.
     * Makes actions:
     * - stops receving any clients requests,
     * - sends unsubscribe messages to clients,
     * - disconnects clients from SIB
     * - stops all modules (handlers, access points).
     * The SIB can be started again with #start() method.
     * @see start()
     */
    void stop();

    BasicOperationHandler* basicOperationHandler_;
    SsapHandler* ssapHandler_;

public slots:
    /**
     * @brief Shutdowns the SIB.
     * Manager shutdowns all modules and unload all plug-ins (access points).
     * After this method SIB cannot be started again.
     */
    void shutdown();


private: // Methods.
    /**
     * @brief Prepares protcols handlers.
     * @see ProtocolHandler
     */
    void prepareProtocolsHandlers();

    /**
     * @brief Prepares operations handlers.
     * Sets mappings between command types and modules (handlers).
     * @see OperationHandler
     * @see SibModule
     * @see CommandType
     */
    void prepareOperationsHandlers();

    /**
     * @brief Finds config file.
     * Try to find config file in paths:
     * 1) given in constructor;
     * 2) current directory (where CuteSIB executable file is located);
     * 3) user home path (in '.cutesib' directory);
     * 4) defined with CUTESIB_CONFIG_PATH in compile-time (e.g. in .pro file).
     * The first path where config file is found will be returned.
     * @return path to config file.
     */
    QString findConfig();


private: // Fields.
    QString configPath_;              ///< Path to config file is given in constructor.
    SibConfig *config_;               ///< SIB configuration.

    AccessPointManager *accessPointManager_;                ///< Manager for access points of the SIB.

    QHash<QString, OperationHandler*> operationHandlers_;   ///< Contains operation handlers by ids.
    QHash<QString, ProtocolHandler*> protocolHandlers_;     ///< Contains protocol handlers by ids.

    Scheduler *scheduler_;              ///< Scheduler of the SIB.
    Triplestore *triplestore_;          ///< Triplestore of the SIB.
    Context context_;                   ///< Context of the SIB.

    QMap<CommandType,QString> operationMapping_;    ///< Mappings between command types and modules.
};

} // namespace cutesib.

#endif // CUTESIB_SIBMANAGER_H
