/**
 * @author Aleksandr A. Lomov <lomov@cs.karelia.ru>
 * @date   09.03.2015
 * @brief  Handler to implement SSAP protocol (SSAP/XML version).
 *
 * This file is part of the CuteSIB project.
 *
 * @section LICENSE
 *
 * CuteSIB is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * CuteSIB is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CuteSIB; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301  USA
 *
 * @section DESCRIPTION
 *
 * Implementation of SSAP/XML protocol for CuteSIB (protocol handler).
 * Supported operation of SSAP/XML protocol:
 * Join/Leave, Insert/Remove/Update/Query (based on triples),
 * Subscription (based on triples)/Unsubscription.
 *
 * Web-site: http://oss.fruct.org/wiki/SmartSpace-platform
 * Copyright (C) 2015-2016 by PetrSU.
 */


#include "ssaphandler.h"

#include <QtCore/QDebug>
#include <QtCore/QString>
#include <QtCore/QThread>

#include "command.h"
#include "scheduler.h"
#include "ssapconstants.h"
#include "ssapmessage.h"
#include "context.h"

#include "tests/stopwatch.h"

#include <QtConcurrent/QtConcurrent>

/*****************************************************************************/
/************************** Constructors/destructors *************************/

// This handler is used for all supported now protocols.
// This is not fully correct but poossible.
cutesib::SsapHandler::SsapHandler(SibConfig *config, QObject *parent) : ProtocolHandler(CUTESIB_SSAPHANDLER_ID,
                                                                     QStringList({CUTESIB_PROTOCOL_SSAPXML,
                                                                                  CUTESIB_PROTOCOL_DISCOVERY,
                                                                                 CUTESIB_PROTOCOL_HTTPSPARQL}), parent)
{
    config_ = config;
}


cutesib::SsapHandler::~SsapHandler()
{

}


/*****************************************************************************/
/****************************** Public methods *******************************/

bool cutesib::SsapHandler::connectAccessPoint(access_point::AccessPointInterface *accessPoint, bool isAsync)
{
    if (accessPoint == nullptr) {
        qWarning() << "Incorrect access point: null pointer.";
        return false;
    }

    // Connect to signals of access potn.
    // With old style it is possible to use signals in interface.
    connect(dynamic_cast<QObject*>(accessPoint),
            SIGNAL(clientConnected(QSharedPointer<cutesib::access_point::Client>)),
            this,
            SLOT(clientConnected(QSharedPointer<cutesib::access_point::Client>)));

    connect(dynamic_cast<QObject*>(accessPoint),
            SIGNAL(clientDisconnected(QSharedPointer<cutesib::access_point::Client>)),
            this,
            SLOT(clientDisconnected(QSharedPointer<cutesib::access_point::Client>)));

    connect(dynamic_cast<QObject*>(accessPoint),
             SIGNAL(requestReceived(QSharedPointer<cutesib::access_point::Request>)),
             this,
             SLOT(requestReceived(QSharedPointer<cutesib::access_point::Request>)));

    if (isAsync) {
        qDebug() << "Asynchronous access point is connect:" << accessPoint->id();
        asyncAccessPoints_.append(accessPoint);
    } else {
        accessPoints_.append(accessPoint);
    }

    return true;
}


void cutesib::SsapHandler::disconnectAccessPoints()
{
    foreach (access_point::AccessPointInterface *accessPoint, accessPoints_) {

        disconnect(dynamic_cast<QObject*>(accessPoint),
                   SIGNAL(clientConnected(QSharedPointer<cutesib::access_point::Client>)),
                   this,
                   SLOT(clientConnected(QSharedPointer<cutesib::access_point::Client>)));

        disconnect(dynamic_cast<QObject*>(accessPoint),
                   SIGNAL(clientDisconnected(QSharedPointer<cutesib::access_point::Client>)),
                   this,
                   SLOT(clientDisconnected(QSharedPointer<cutesib::access_point::Client>)));

        disconnect(dynamic_cast<QObject*>(accessPoint),
                   SIGNAL(requestReceived(QSharedPointer<cutesib::access_point::Request>)),
                   this,
                   SLOT(requestReceived(QSharedPointer<cutesib::access_point::Request>)));
    }

    accessPoints_.clear();

    foreach (access_point::AccessPointInterface *accessPoint, asyncAccessPoints_) {

        disconnect(dynamic_cast<QObject*>(accessPoint),
                   SIGNAL(clientConnected(QSharedPointer<cutesib::access_point::Client>)),
                   this,
                   SLOT(clientConnected(QSharedPointer<cutesib::access_point::Client>)));

        disconnect(dynamic_cast<QObject*>(accessPoint),
                   SIGNAL(clientDisconnected(QSharedPointer<cutesib::access_point::Client>)),
                   this,
                   SLOT(clientDisconnected(QSharedPointer<cutesib::access_point::Client>)));

        disconnect(dynamic_cast<QObject*>(accessPoint),
                   SIGNAL(requestReceived(QSharedPointer<cutesib::access_point::Request>)),
                   this,
                   SLOT(requestReceived(QSharedPointer<cutesib::access_point::Request>)));
    }

    asyncAccessPoints_.clear();
}


bool cutesib::SsapHandler::start()
{
    if (accessPoints_.size() == 0) {
        qDebug() << "No access points connected connect to handler.";
        return false;
    }

    foreach (access_point::AccessPointInterface *accessPoint, accessPoints_) {
        if (accessPoint->start() != 0) {
            qDebug() << "Access point "<<  accessPoint->id() << "cannot be started.";
        }
    }

    foreach (access_point::AccessPointInterface *accessPoint, asyncAccessPoints_) {

        QThread* thread = new QThread();

        dynamic_cast<QObject*>(accessPoint)->moveToThread(thread);
        connect(thread, SIGNAL(started()), dynamic_cast<QObject*>(accessPoint), SLOT(start()));
        connect(this, SIGNAL(stopping()), dynamic_cast<QObject*>(accessPoint), SLOT(stop()));

        connect(dynamic_cast<QObject*>(accessPoint), SIGNAL(stopped()), thread, SLOT(quit()));
        connect(thread, SIGNAL(finished()), thread, SLOT(deleteLater()));

        thread->start();
    }

    return true;
}


void cutesib::SsapHandler::stop()
{
    qDebug() << id();

    foreach(auto client, subscribtionIds_.keys()) {

        QString id = subscribtionIds_.value(client);

        CommandRequest *request = new CommandRequest(CommandType::UNSUBSCRIBE, this->id());

        request->setConfirmation(false);

        request->setSubscriptionId(id);

        context()->scheduler()->perform(request);

        ///TODO: space id.
        ///Send unsubscribe.
        //SsapMessage unsubscribe(SsapTransactionType::UNSUBSCRIBE, SsapMessageType::REQUEST, client->nodeId(), "X", id, "1");

        //client->send(unsubscribe.createConfirmation())
    }

    subscribtionIds_.clear();

    foreach (access_point::AccessPointInterface *accessPoint, accessPoints_) {
        accessPoint->stop();
    }

    emit stopping();
}


void cutesib::SsapHandler::shutdown()
{
    qDebug() << id();
    disconnectAccessPoints();
}


cutesib::CommandResponse *cutesib::SsapHandler::perform(cutesib::CommandRequest* command)
{
    if (command->Type() != CommandType::SUBSCRIPTION_INDICATE && command->needConfirmation()) {
#if __GNUC__ < 5 && __GNUC_MINOR__ < 7
        qWarning() << "Incorrect command:" << static_cast<int>(command->Type());
#else
        qWarning() << "Incorrect command:" << static_cast<std::underlying_type<CommandType>::type>(command->Type());
#endif
        return new CommandResponse(command, SibError::NOT_IMPLEMENTED);

    } else if (command->Type() != CommandType::SUBSCRIPTION_INDICATE && !command->needConfirmation()) {
#if __GNUC__ < 5 && __GNUC_MINOR__ < 7
        qWarning() << "Incorrect command:" << static_cast<int>(command->Type());
#else
        qWarning() << "Incorrect command:" << static_cast<std::underlying_type<CommandType>::type>(command->Type());
#endif
        delete command;
        return nullptr;
    }

    foreach (auto client, subscribtionIds_.keys()) {
        if (subscribtionIds_.value(client) == command->subscriptionId()) {
            qDebug() << "Send to client" << "All clients: " << subscribtionIds_.count();

            SsapMessage message(SsapTransactionType::SUBSCRIBE, SsapMessageType::INDICATION, client->nodeId(), "X",
                                command->subscriptionId(), "1");

            if (command->insertTriples().isEmpty() && command->removeTriples().isEmpty()) {
                message.setSubscriptionNewResults(
                            command->getParameter(CommandParameter::SUBSCRIPTION_NEW_RESULT).toString());
                message.setSubscriptionOldResults(
                            command->getParameter(CommandParameter::SUBSCRIPTION_OLD_RESULT).toString());
            } else {
                message.setResultInsertTriples(command->insertTriples());
                message.setResultRemoveTriples(command->removeTriples());
            }


            int result = client->send(message.createIndication());
            bool commandResult = (result == 0) ? true : false;

            if (command->needConfirmation()) {
                CommandResponse * response = new CommandResponse(command, commandResult);
                return response;
            }
        }
    }

    delete command;
    return nullptr;
}


/*****************************************************************************/
/******************************* Public slots ********************************/

void cutesib::SsapHandler::clientDisconnected(QSharedPointer<cutesib::access_point::Client> client)
{
    QString subscriptionId = subscribtionIds_.take(client);

    if (!subscriptionId.isEmpty()) {

        CommandRequest *command = new CommandRequest(CommandType::UNSUBSCRIBE, this->id(),false);
        command->setSubscriptionId(subscriptionId);

        context()->scheduler()->perform(command);
    }
}


void cutesib::SsapHandler::clientConnected(QSharedPointer<cutesib::access_point::Client> client)
{
    qDebug() << "Client connected from " << client->getAccessPoint()->id();
}


void cutesib::SsapHandler::requestReceived(QSharedPointer<cutesib::access_point::Request> request)
{
    Stopwatch::start();

    SsapMessage* message = new SsapMessage();
    message->load(request->data());

    Stopwatch::elapsed();

    qDebug() << "SSAP message\n"
             << "\tTransaction type:" << static_cast<int>(message->transactionType()) << message->transactionTypeText() << "\n"
             << "\tMessage type:" << message->messageType() << "\n"
             << "\tParameters type:" << message->parameterType() << "\n"
             << "\tParameters query:" << message->parameterQuery() << "\n"
             << "\tNeed confirmation" << message->isNeedConfirmation() << "\n"
             << "\tTransaction id:" << message->transactionId() << "\n"
             << "\tSubscription id:" << message->subscriptionId() << "\n"
             << "\tNode id:" << message->nodeId() << "\n"
             << "\tSpace id:" << message->spaceId() << "\n"
             << "\tInsert triples" << message->insertTriples().count() << "\n"
             << "\tRemove triples" << message->removeTriples().count() << "\n"
             << "\tQuery triples" << message->queryTriples().count();



    handleMessage(request, message);
}

void cutesib::SsapHandler::handleParallelQueryResponse(cutesib::CommandResponse *commandResponse)
{

    if (commandResponse == NULL) {
        qWarning() << "commandResponse is NULL";
        return;
    }
    else if (commandResponse == nullptr) {
        qWarning() << "commandResponse is nullptr";
        return;
    }

    CommandRequest* commandRequest = commandResponse->getRequest();

    if (commandRequest == NULL) {
        qWarning() << "commandRequest is NULL";
        return;
    }

    cutesib::SsapMessage* message = commandRequest->ssapMessage;
    QSharedPointer<cutesib::access_point::Request> request = commandRequest->request;

    if (!commandResponse->triples().isEmpty()) {
        message->setResultQueryTriples(commandResponse->triples());
    } else {
        message->setQueryResults(commandResponse->getParameter(CommandParameter::QUERY_RESULT_TEXT).toString());
    }

    if (commandResponse == nullptr) {
        return;
    }

    if (commandResponse->isSuccessful()) {
        //request->client()->send(message->createConfirmation());
        QtConcurrent::run(request->client().data(), &Client::send, message->createConfirmation());
    } else {
        //request->client()->send(message->createErrorConfirmation());
        QtConcurrent::run(request->client().data(), &Client::send, message->createErrorConfirmation());
    }

    if (message->transactionType() == SsapTransactionType::UNSUBSCRIBE) {
       request->client()->close();
    }

    delete commandResponse;
}


/*****************************************************************************/
/***************************** Private methods *******************************/

void cutesib::SsapHandler::handleMessage(QSharedPointer<cutesib::access_point::Request> request, SsapMessage* message)
{
    if (message->messageType() != SsapConstants::MessageTypeRequest) {
        qDebug() << "Incorrect message: message type is not a request.";
        request->client()->send(message->createErrorConfirmation());
        return;
    }

    CommandResponse *commandResponse = nullptr;

    // Handle message by type.
    // 'Join' and 'Leave' do nothing now - send only confirmation.
    // 'Leave' must closes all subscription Now this done by disconnection signal from access point.
    if (message->transactionType() == SsapTransactionType::JOIN) {

        request->client()->setNodeId(message->nodeId());
        request->client()->send(message->createConfirmation());

    } else if (message->transactionType() == SsapTransactionType::LEAVE) {

        request->client()->setNodeId(QString::null);
        request->client()->send(message->createConfirmation());

    } else if (message->transactionType() == SsapTransactionType::INSERT) {

        commandResponse = handleInsertMessage(message);

    } else if (message->transactionType() == SsapTransactionType::QUERY) {

        commandResponse = handleQueryMessage(request, message);

    } else if (message->transactionType() == SsapTransactionType::UPDATE) {

        commandResponse = handleUpdateMessage(message);

    } else if (message->transactionType() == SsapTransactionType::REMOVE) {

        commandResponse = handleRemoveMesage(message);

    } else if (message->transactionType() == SsapTransactionType::SUBSCRIBE) {

        commandResponse = handleSubscribeMessage(request, message);

    } else if (message->transactionType() == SsapTransactionType::UNSUBSCRIBE) {
        commandResponse = handleUnsubscribeMessage(request, message);

    }

    if (commandResponse == nullptr) {
        return;
    }

    if (commandResponse->isSuccessful()) {
        request->client()->send(message->createConfirmation());
    } else {
        request->client()->send(message->createErrorConfirmation());
    }

    if (message->transactionType() == SsapTransactionType::UNSUBSCRIBE) {
       request->client()->close();
    }

    delete commandResponse;
}


cutesib::CommandResponse* cutesib::SsapHandler::handleInsertMessage(SsapMessage *message)
{
    CommandRequest *command = new CommandRequest(CommandType::INSERT, this->id());

    QList<cutesib::Triple *> triples = message->insertTriples();

    command->setInsertTriples(triples);

    command->setConfirmation(message->isNeedConfirmation());

    return context()->scheduler()->perform(command);
}


cutesib::CommandResponse *cutesib::SsapHandler::handleQueryMessage(QSharedPointer<cutesib::access_point::Request> request, SsapMessage* message)
{
    CommandType type = (message->isSparqlQuery()) ? CommandType::SPARQL_QUERY : CommandType::QUERY;

    CommandRequest *command = new CommandRequest(type, this->id());

    // SSAP Query is always confirmed.
    command->setConfirmation(true);

    command->setQueryTriples(message->queryTriples());

    command->setParameter(CommandParameter::QUERY_TEXT, message->parameterQuery());
    command->ssapMessage = message;
    command->request = request;

    // PARALLEL OR SEQUENTIAL QUERY
    if (config_->getTriplestoreParameter("parallel") == "query") {
        Scheduler* scheduler = context()->scheduler();
        QtConcurrent::run(scheduler, &Scheduler::perform, command);
        return nullptr;
    } else {
        CommandResponse *commandResponse = context()->scheduler()->perform(command);

        if (!commandResponse->triples().isEmpty()) {
            message->setResultQueryTriples(commandResponse->triples());
        } else {
            message->setQueryResults(commandResponse->getParameter(CommandParameter::QUERY_RESULT_TEXT).toString());
        }

        return commandResponse;
    }
}


cutesib::CommandResponse* cutesib::SsapHandler::handleUpdateMessage(SsapMessage *message)
{
    CommandRequest *command = new CommandRequest(CommandType::UPDATE, this->id());

    command->setConfirmation(message->isNeedConfirmation());

    command->setInsertTriples(message->insertTriples());
    command->setRemoveTriples(message->removeTriples());

    return context()->scheduler()->perform(command);
}


cutesib::CommandResponse* cutesib::SsapHandler::handleRemoveMesage(SsapMessage *message)
{
    CommandRequest *command = new CommandRequest(CommandType::REMOVE, this->id());

    command->setConfirmation(message->isNeedConfirmation());

    command->setRemoveTriples(message->removeTriples());

    return context()->scheduler()->perform(command);
}


cutesib::CommandResponse* cutesib::SsapHandler::handleSubscribeMessage(QSharedPointer<cutesib::access_point::Request> request, SsapMessage *message)
{
    CommandRequest *command = new CommandRequest(CommandType::SUBSCRIBE, this->id());

    // SSAP subscribe is always confirmed.
    command->setConfirmation(message->isNeedConfirmation());

    command->setQueryTriples(message->queryTriples());

    command->setParameter(CommandParameter::QUERY_TEXT, message->parameterQuery());

    QString subscriptionId = createSubscriptionId(message);
    command->setSubscriptionId(subscriptionId);

    CommandResponse *commandResponse = context()->scheduler()->perform(command);

    if (commandResponse->isSuccessful()) {

        if (!commandResponse->triples().isEmpty()) {
            message->setResultQueryTriples(commandResponse->triples());
        } else {
            message->setQueryResults(commandResponse->getParameter(CommandParameter::SUBSCRIPTION_NEW_RESULT).toString());
        }

        auto client = request->client();

        client->setNodeId(message->nodeId());

        subscribtionIds_.insert(client, subscriptionId);
        request->accessPoint()->startChecking(client);

        qDebug() << "Count of subscriptions:" << subscribtionIds_.values();
    }

    return commandResponse;
}


cutesib::CommandResponse* cutesib::SsapHandler::handleUnsubscribeMessage(QSharedPointer<Request> request, SsapMessage *message)
{
    CommandRequest *command = new CommandRequest(CommandType::UNSUBSCRIBE, this->id());
    command->setConfirmation(message->isNeedConfirmation());
    command->setSubscriptionId(message->subscriptionId());

    CommandResponse *commandResponse = context()->scheduler()->perform(command);

    qDebug() << "Try to found subscriber:" << message->subscriptionId() << ". Subscribers:" << subscribtionIds_.count();
    foreach (auto client, subscribtionIds_.keys()) {
        if (subscribtionIds_.value(client) == message->subscriptionId()) {
            request->accessPoint()->stopChecking(client);
            //request->getAccessPoint()->send(client, response);
            subscribtionIds_.remove(client);
            qDebug() << "Subscriber removed:" << message->subscriptionId()  << ". Subscribers:" << subscribtionIds_.count();
            break;
        }
    }

    return commandResponse;
}


QString cutesib::SsapHandler::createSubscriptionId(SsapMessage *message) const
{
    // Postfix index for subscription id.
    static int index = 1;

    message->setSubscriptionId(message->nodeId() + "_" + QString::number(index));

    ++index;

    return message->subscriptionId();
}
