/**
 * @author Aleksandr A. Lomov <lomov@cs.karelia.ru>
 * @date   09.03.2015
 * @brief  Handler to implement SSAP protocol (SSAP/XML version).
 *
 * This file is part of the CuteSIB project.
 *
 * @section LICENSE
 *
 * CuteSIB is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * CuteSIB is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CuteSIB; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301  USA
 *
 * @section DESCRIPTION
 *
 * Implementation of SSAP/XML protocol for CuteSIB (protocol handler).
 * Supported operation of SSAP/XML protocol:
 * Join/Leave, Insert/Remove/Update/Query (based on triples),
 * Subscription (based on triples)/Unsubscription.
 *
 * Web-site: http://oss.fruct.org/wiki/SmartSpace-platform
 * Copyright (C) 2015-2016 by PetrSU.
 */


#ifndef CUTESIB_SSAPHANDLER_H
#define CUTESIB_SSAPHANDLER_H

#include <QObject>

#include "client.h"
#include "accesspointinterface.h"

#include "ssapmessage.h"
#include "sibconfig.h"
#include "client.h"

#include "command.h"
#include "commandresponse.h"
#include "protocolhandler.h"

#include "sibconfig.h"


#define CUTESIB_SSAPHANDLER_ID "PH-SSAP-XML"

namespace cutesib {

/**
 * @brief The SsapHandler class implements SSAP/XML protocol.
 * Handler works with access points using signals/slots.
 * Each access point process connection/disconnection of agent, receiving of requests and sending responses.
 *
 * Handler ID: PH:SSAP-XML
 */
class SsapHandler : public ProtocolHandler
{
    Q_OBJECT

public:
    /**
     * @brief SsapHandler constructor.
     * @param config for handler.
     * @param parent of instance.
     */
    explicit SsapHandler(SibConfig *config, QObject *parent = 0);

    /**
     * @brief SsapHandler destructor.
     */
    ~SsapHandler();

    /**
     * @brief Connects access point to handler.
     * Access point process connection with agent: receives requests and sends response.
     * Several access points can be connected to handler.
     * @param accessPoint to connect.
     * @param isAsync flag to use access point in main (SIB) thread or other thread.
     * @see AccessPointInterface
     */
    bool connectAccessPoint(access_point::AccessPointInterface *accessPoint, bool isAsync) /*override*/;

    /**
     * @brief Disconnects all access points from handler.
     */
    void disconnectAccessPoints() /*override*/;

    /**
     * @brief Starts the handler.
     * Handler starts to process request from access points.
     * @return true on success or false otherwise.
     */
    bool start() /*override*/;

    /**
     * @brief Stop the handler.
     * Handler stops to process request from access points.
     */
    void stop() /*override*/;

    /**
     * @brief Shutdowns the handler.
     * Stops all activity of handler, disconnects all access points and
     * prepares handler for clean deletion.
     */
    void shutdown() /*override*/;

    /**
     * @brief Performs a command with operation.
     * Handler performs only command to send indication by subscriptions.
     * @param command with operation to perform.
     * @return response for command with a result of performing operation
     * if the command does not requared response then returns NULL.
     * @see Command
     * @see CommandRequest
     * @see CommandResponse
     */
    CommandResponse* perform(CommandRequest *command) /*override*/;

    /**
     * @brief Gets protocols supported by handler.
     * Handler supports SSAP/XML protocol.
     * @return list with protocols id/names.
     */
    QStringList protocols() { return QStringList(CUTESIB_PROTOCOL_SSAPXML); }


public slots: // Slots for access points.
    /**
     * @brief Receives signal when agent disconnect from access point.
     * @param client that was disconnected.
     */
    void clientDisconnected(QSharedPointer<cutesib::access_point::Client> client);

    /**
     * @brief Receives signal when agent connect to access point.
     * @param client that was connected.
     */
    void clientConnected(QSharedPointer<cutesib::access_point::Client> client);

    /**
     * @brief Receives signal with requets from agent.
     * @param request from agent than was received.
     */
    void requestReceived(QSharedPointer<cutesib::access_point::Request> request);

    void handleParallelQueryResponse(CommandResponse* commandResponse);

signals:
    /**
     * @brief Signals threads and asynchronous access points to stop working.
     */
    void stopping();


private:    // Methods.
    /**
     * @brief Handler SSAP/XML message that was received from agent.
     * @param request with raw message of agent.
     * @param message in SIB-native format.
     */
    void handleMessage(QSharedPointer<Request> request, SsapMessage *message);

    // Methods to handler messages by type.
    cutesib::CommandResponse* handleInsertMessage(SsapMessage* message);
    cutesib::CommandResponse* handleQueryMessage(QSharedPointer<Request> request, SsapMessage *message);
    cutesib::CommandResponse* handleUpdateMessage(SsapMessage *message);
    cutesib::CommandResponse* handleRemoveMesage(SsapMessage *message);
    cutesib::CommandResponse* handleSubscribeMessage(QSharedPointer<Request> request, SsapMessage *message);
    cutesib::CommandResponse *handleUnsubscribeMessage(QSharedPointer<Request> request, SsapMessage *message);

    /**
     * @brief Creates identifier for subscription.
     * Identifier is created based on node id from SSAP message and index: nodeId + _ + index;
     * @param message of SSAP SUBSCRIBE.
     * @return identifier for subscription.
     */
    QString createSubscriptionId(SsapMessage *message) const;


private:    // Fields
    /**
     * @brief Access points of handler.
     * Each access points communicate with handler by signals/slots.
     * Slots connects with signals when handler starts and disconnect when handler stops working.
     * @see AccessPointInterface
     */
    QList<cutesib::access_point::AccessPointInterface*> accessPoints_;

    /**
     * @brief Asynchronous access points of handler.
     * Asynchronous access points works in own threads.
     * @see AccessPointInterface
     * @see accessPoints_
     */
    QList<cutesib::access_point::AccessPointInterface*> asyncAccessPoints_;

    /// TODO: Works with subscription, use another container.
    /**
     * @brief Map for agents and their subscription indentifications.
     */
    QMap<QSharedPointer<cutesib::access_point::Client>, QString> subscribtionIds_;

    SibConfig *config_;
};

} // namespace cutesib

#endif // CUTESIB_SSAPHANDLER_H
