/**
 * @author Aleksandr A. Lomov <lomov@cs.karelia.ru>
 * @date   13.04.2015
 * @brief  Describes a SSAP message.
 *
 * @section LICENSE
 *
 * CuteSIB is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * CuteSIB is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CuteSIB; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301  USA
 *
 * @section DESCRIPTION
 *
 * This class is used as a container for SSAP/XML message data.
 * It contains all needed data to handle request message from client
 * and can be used to create a response message.
 *
 * This file is part of the CuteSIB project.
 *
 * Copyright (C) CuteSIB team (PetrSU). All rights reserved.
 * Contact: http://oss.fruct.org/wiki/SmartSpace-platform
 */


#ifndef CUTESIB_SSAPMESSAGE_H
#define CUTESIB_SSAPMESSAGE_H


#include "triple.h"

#include <QtCore/QObject>
#include <QtCore/QList>
#include <QtCore/QString>
#include <QtCore/QXmlStreamReader>


namespace cutesib {

/** @brief Transaction types of SSAP message. */
enum class SsapTransactionType {
    UNKNOWN = -1,      /**< Wrong/unknown type of SSAP message. */
    JOIN,              /**< Describes joining to the smart space. */
    LEAVE,             /**< Describes leaving to the smart space. */
    INSERT,            /**< Describes inserting data to the smart space. */
    UPDATE,            /**< Describes updating data in the smart space. */
    REMOVE,            /**< Describes removing data from the smart space. */
    QUERY,             /**< Describes quering data from the smart space. */
    SUBSCRIBE,         /**< Describes subscribing to data in the smart space. */
    UNSUBSCRIBE        /**< Describes unsubscribing from data in the smart space. */
};

/** @brief Message types of SSAP messages. */
enum class SsapMessageType {
    REQUEST,           /**< Request type of message (sent by agents to SIB). */
    CONFIRME,          /**< Confirmation type of message to requests (sent from SIB to agents). */
    INDICATION         /**< Indication about changes by subscription (sent from SIB to agents). */
};


/**
 * @brief The SsapMessage class
 */
class SsapMessage : public QObject
{
    Q_OBJECT
public:
    /**
     * @brief SsapMessage constructor.
     * @param parent of the instance.
     */
    explicit SsapMessage(QObject *parent = 0);

    /**
     * @brief SsapMessage class constrructor with all main parameters for SSAP message.
     * @param type.
     * @param messageType of the message.
     * @param nodeId.
     * @param spaceId.
     * @param subscriptionId.
     * @param transactionId.
     * @param parent of the instance.
     */
    explicit SsapMessage(SsapTransactionType parameterType, SsapMessageType messageType,
                         QString nodeId, QString spaceId, QString subscriptionId = QString::null, QString transactionId = QString::null,
                         QObject *parent = 0);

    /**
     * @brief SsapMessage destructor.
     */
    ~SsapMessage();

    /**
     * @brief Loads SSAP/XML message to parse it.
     * @param message SSAP message to parse it.
     * @return true on success ot false otherwise (incorrect message).
     */
    bool load(QByteArray message);

    /**
     * @brief Gets a transaction type of the message.
     * @return transaction type.
     * @see SsapTransactionType
     */
    SsapTransactionType transactionType() const;

    /**
     * @brief Gets a string representation of the transaction type.
     * @return transaction type in the text form.
     */
    QString transactionTypeText() const;

    /**
     * @brief Gets a message type (request/response).
     * @return message type.
     */
    QString messageType();

    /**
     * @brief Gets a node identifier from the message.
     * @return node identitfier.
     */
    QString nodeId() const;

    /**
     * @brief Gets a smart space identifier from the message.
     * @return smart space identifier.
     */
    QString spaceId() const;

    /**
     * @brief Gets a truncsaction identitfer from the message.
     * @return transaction identifier.
     */
    QString transactionId() const;


    /**
     * @brief Gets an identifier of subscription.
     * @return identifier of subscription.
     */
    QString subscriptionId() const;

    /**
     * @brief Gets a value of type parameter.
     * Parameter: <parameter name = "type">...</parameter>
     * @return value of parameter.
     */
    QString parameterType() const;

    /**
     * @brief Gets a value of query parameter.
     * Parameter: <parameter name = "query">...</parameter>
     * @return value of parameter.
     */
    QString parameterQuery() const;

    /**
     * @brief Checks: message is SPARQL query.
     * @return true if messsage contains SPARQL query or false otherwise.
     */
    bool isSparqlQuery() const;

    /**
     * @brief Sets subscription identifier.
     * Identifier is used to create SSAP SUBSCRIBE confirmation message.
     * @param id is identifier of subscription.
     */
    void setSubscriptionId(QString id);

    /**
     * @brief Creates a response based current information.
     * @return bytes of SSAP/XML response.
     */
    QByteArray createConfirmation() const;

    /**
     * @brief Creates a error response based current information.
     * @return bytes of SSAP/XML error response.
     */
    QByteArray createErrorConfirmation() const;

    /**
     * @brief Creates an indication for subscription based current information.
     * @return bytes of SSAP/XML indication.
     */
    QByteArray createIndication() const;

    /**
     * @brief Flag: message requires a response or not.
     * @return true if the response required or false otherwise.
     */
    bool isNeedConfirmation() const;

    // TODO: use shared pointer.
    // Getters/Setters for triples.
    const QList<Triple *> insertTriples() const;
    const QList<Triple *> removeTriples() const;
    const QList<Triple *> queryTriples() const;

    void setResultInsertTriples(QList<Triple *> triples);
    void setResultRemoveTriples(QList<Triple *> triples);
    void setResultQueryTriples(QList<Triple *> triples);

    void setQueryResults(QString queryResults) { queryResults_ = queryResults; }

    // For SPARQL subscription.
    void setSubscriptionNewResults(QString subscriptionResults) { subscriptionNewResults_ = subscriptionResults; }
    void setSubscriptionOldResults(QString subscriptionResults) { subscriptionOldResults_ = subscriptionResults; }

private: // Methods
    SsapTransactionType convertTrsansactionType(QString transactionType);
    void handleParameter(QXmlStreamReader &doc);
    QList<Triple*> handleTripleList(QXmlStreamReader &doc);
    Triple *handleTriple(QXmlStreamReader &doc);
    QString getAttribute(QXmlStreamReader &doc, QString name);
    RdfTripleType getRdfType(QXmlStreamReader &doc);
    static QString triplesToSsapList(QList<Triple *> triples);
    void readNotNullNext();

private: // Fields
    QXmlStreamReader* xmlDoc;
    SsapTransactionType transactionType_;
    QString transactionTypeText_;
    QString messageType_;
    QString transactionId_;
    QString nodeId_;
    QString spaceId_;
    QString subscriptionId_;
    QString type_; // parameter type
    QString query_; // Parameter query

    bool needConfirmation_ /*= true*/;

    QList<Triple*> insertTriples_;
    QList<Triple*> removeTriples_;
    QList<Triple*> queryTriples_;

    QList<Triple*> resultInsertTriples_;
    QList<Triple*> resultRemoveTriples_;
    QList<Triple*> resultQueryTriples_;
    QString queryResults_; ///< To store results of non-triple query operation (now only for SPARQL results).
    QString subscriptionNewResults_; ///< To store results of non-triple subscription operation (now only for SPARQL results).
    QString subscriptionOldResults_; ///< To store results of non-triple subscription operation (now only for SPARQL results).

};


}
#endif // CUTESIB_SSAPMESSAGE_H

