/**
* @author Ivan V. Galov <galov@cs.karelia.ru>
* @date   18.04.2016
* @brief  Class containing information about one subscription.
*
* This file is part of the CuteSIB project.
*
* @section LICENSE
*
* CuteSIB is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* CuteSIB is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with CuteSIB; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor,
* Boston, MA  02110-1301  USA
*
* @section DESCRIPTION
*
* Class containing information about one subscription.
*
* Web-site: http://oss.fruct.org/wiki/SmartSpace-platform
* Copyright (C) 2015-2016 by PetrSU.
**/

#include "subscription.h"
#include "commandrequest.h"

#include <QDebug>

cutesib::Subscription::Subscription(QString subscriptionId, QList<Triple*> queryList)
{
    subscriptionId_ = subscriptionId;
    subworld_ = librdf_new_world();
    librdf_world_open(subworld_);

    // TODO: add ifdef (old libs of redland)
    //rasqal_world_set_warning_level(librdf_world_get_rasqal(subworld_), 0);

    /*if (p->subs_hash)
           substorage=librdf_new_storage(rsp->subworld, "hashes", NULL, "hash-type='memory'");
    else
           substorage=librdf_new_storage(rsp->subworld,  NULL, NULL, NULL);
           */
    // TODO: persistent storages for subscriptions?

    substorage_ = librdf_new_storage(subworld_, "hashes", NULL, "hash-type='memory'");

    submodel_ = librdf_new_model(subworld_, substorage_, NULL);

    queryList_ = queryList;
}

cutesib::Subscription::~Subscription()
{
    foreach(Triple* triple, queryList_) {
        delete triple;
    }

    librdf_free_model(submodel_);
    librdf_free_storage(substorage_);
    librdf_free_world(subworld_);
}

QString cutesib::Subscription::getSubscriptionId()
{
    return subscriptionId_;
}

void cutesib::Subscription::checkSubscription(QList<cutesib::Triple *> insertedTriplesList, QList<cutesib::Triple *> removedTriplesList, SubscriptionIndication& indication)
{
    bool same_s;
    bool same_p;
    bool same_o;

    bool result = false;

    QList<cutesib::Triple *> matchedInsertedTriplesList;
    QList<cutesib::Triple *> matchedRemovedTriplesList;

    foreach(Triple* queryTriple, queryList_)
    {
        /*
        //////////////////FILTER OUT IDENTICAL QUERY PATTERNS//////////////////////
        if (op->rsp->sparql_subscribe)
        {
            if (tq->replicated)
            {
                query_list=query_list->next;
                continue;
            }
        }

        //////////////////END FILTER IDENTICAL QUERY PATTERNS//////////////////////
        */

        foreach(Triple* removedTriple, removedTriplesList) {
            same_s = false;
            same_p = false;
            same_o = false;

            if (
                    (removedTriple->subject() == queryTriple->subject()) ||
                    (queryTriple->subject() == Triple::AnyUri) ||
                    (queryTriple->subject() == Triple::AnyShortForm)
            )
            {
                same_s = true;
            }
            else
            {
                continue;
            }

            if (
                    (removedTriple->predicate() == queryTriple->predicate()) ||
                    (queryTriple->predicate() == Triple::AnyUri) ||
                    (queryTriple->predicate() == Triple::AnyShortForm)
            )
            {
                same_p = true;
            }
            else
            {
                continue;
            }


            if (
                    ((removedTriple->object() == queryTriple->object()) && (removedTriple->objectType() == queryTriple->objectType())) ||
                    (queryTriple->object() == Triple::AnyUri) ||
                    (queryTriple->object() == Triple::AnyShortForm)
            )
            {
                same_o = true;
            }
            else
            {
                continue;
            }

            if (same_s && same_p && same_o) {
                //No wildcard on statement removed
                // ssTriple_t * copied_triple = ssTripleHardCopy(removed);

                matchedRemovedTriplesList.append(removedTriple);
                result = true;
            }
        }


        foreach(Triple* insertedTriple, insertedTriplesList) {
            same_s = false;
            same_p = false;
            same_o = false;

            if (
                    (insertedTriple->subject() == queryTriple->subject()) ||
                    (queryTriple->subject() == Triple::AnyUri) ||
                    (queryTriple->subject() == Triple::AnyShortForm)
            )
            {
                same_s = true;
            }
            else
            {
                continue;
            }

            if (
                    (insertedTriple->predicate() == queryTriple->predicate()) ||
                    (queryTriple->predicate() == Triple::AnyUri) ||
                    (queryTriple->predicate() == Triple::AnyShortForm)
            )
            {
                same_p = true;
            }
            else
            {
                continue;
            }


            if (
                    ((insertedTriple->object() == queryTriple->object()) && (insertedTriple->objectType() == queryTriple->objectType())) ||
                    (queryTriple->object() == Triple::AnyUri) ||
                    (queryTriple->object() == Triple::AnyShortForm)
            )
            {
                same_o = true;
            }
            else
            {
                continue;
            }

            if (same_s && same_p && same_o) {
                //No wildcard on statement removed
                // ssTriple_t * copied_triple = ssTripleHardCopy(added);

                matchedInsertedTriplesList.append(insertedTriple);
                result = true;
            }
        }

    }

    /************************************************/

    librdf_statement * statement_added;
    librdf_statement * statement_removed;

    QList<librdf_statement*> addedStatementsList;
    QList<librdf_statement*> removedStatementsList;

    QList<Triple*> insertedIndicationTriplesList;
    QList<Triple*> removedIndicationTriplesList;

    foreach (Triple* removedTriple, matchedRemovedTriplesList) {
        ////////CHECKIG FOR IDENTICAL TRIPLES INSERTED AND REMOVED IN THE SAME UPDATE///
        bool found_in_added = false;

        foreach (Triple* insertedTriple, matchedInsertedTriplesList) {
            if (
                 (insertedTriple->subject() == removedTriple->subject()) &
                 (insertedTriple->predicate() == removedTriple->predicate()) &
                 (insertedTriple->object() == removedTriple->object()) &
                 (insertedTriple->objectType() == removedTriple->objectType())
                )
            {
                found_in_added = true;
                break;
            }
        }

        if (!found_in_added)
        {
            statement_removed = librdf_new_statement(subworld_);
            librdf_statement_set_subject(statement_removed, librdf_new_node_from_uri_string(subworld_, (unsigned char*) removedTriple->subject().toUtf8().data()));
            librdf_statement_set_predicate(statement_removed, librdf_new_node_from_uri_string(subworld_, (unsigned char*) removedTriple->predicate().toUtf8().data()));

            if (removedTriple->objectType() == RdfTripleType::URI)
            {
                librdf_statement_set_object(statement_removed,
                        librdf_new_node_from_uri_string(subworld_, (unsigned char*) removedTriple->object().toUtf8().data()));
            }
            else
            {
                librdf_statement_set_object(statement_removed,
                        librdf_new_node_from_literal(subworld_, (unsigned char*) removedTriple->object().toUtf8().data(), NULL, 0));
            }


            librdf_stream * stream_tmp = NULL;
            stream_tmp = librdf_model_find_statements(submodel_, statement_removed);

            //If the triple is present, it will be removed for sure.
            //REDLAND BUG; USING STREAM
            if (!librdf_stream_end(stream_tmp))
            {
                // really removed
                removedStatementsList.append(librdf_new_statement_from_statement(statement_removed));
                removedIndicationTriplesList.append(removedTriple->createCopy());
            }
            librdf_free_statement(statement_removed);
            librdf_free_stream(stream_tmp);
        }
    }


    foreach (Triple* insertedTriple, matchedInsertedTriplesList) {

        statement_added = librdf_new_statement(subworld_);

        librdf_statement_set_subject(statement_added, librdf_new_node_from_uri_string(subworld_, (unsigned char*) insertedTriple->subject().toUtf8().data()));
        librdf_statement_set_predicate(statement_added, librdf_new_node_from_uri_string(subworld_, (unsigned char*) insertedTriple->predicate().toUtf8().data()));

        if (insertedTriple->objectType() == RdfTripleType::URI)
        {
            librdf_statement_set_object(statement_added,
                    librdf_new_node_from_uri_string(subworld_, (unsigned char*) insertedTriple->object().toUtf8().data()));
        }
        else
        {
            librdf_statement_set_object(statement_added,
                    librdf_new_node_from_literal(subworld_, (unsigned char*) insertedTriple->object().toUtf8().data(), NULL, 0));
        }

        //If the triple is not present, it will be added for sure.
        if (librdf_model_contains_statement(submodel_, statement_added) == 0)
        {
            // really removed
            addedStatementsList.append(librdf_new_statement_from_statement(statement_added));
            insertedIndicationTriplesList.append(insertedTriple->createCopy());
        }

        librdf_free_statement(statement_added);

    }

    /************************************************/

    //add_triplelist_in_submodel    (p, op->rsp, op->rsp->added);
    foreach (librdf_statement* statement, addedStatementsList) {
        librdf_model_add_statement(submodel_, statement);
        librdf_free_statement(statement);
    }

    //remove_triplelist_in_submodel (p, op->rsp, op->rsp->removed);
    foreach (librdf_statement* statement, removedStatementsList) {
        librdf_model_remove_statement(submodel_, statement);
        librdf_free_statement(statement);
    }

    //if( ++(op->rsp->ind_seqnum) == SSAP_IND_WRAP_NUM )
    //    op->rsp->ind_seqnum=1;

    qDebug(" ********** SUBSCRIPTION IND!\n");

    if(!insertedIndicationTriplesList.empty()) {
        indication.insertedTriples = insertedIndicationTriplesList;
    }

    if(!removedIndicationTriplesList.empty()) {
        indication.removedTriples = removedIndicationTriplesList;
    }


    /*
    CommandRequest* indicationRequest = new CommandRequest(CommandTypes::COMMAND_SUBSCRIBE,
                                                           SibModule::MODULE_SUBSCRIPTION_HANDLER,
                                                           SibModule::MODULE_SUBSCRIPTION_HANDLER);
    indicationRequest->setInsertTriples(insertedIndicationTriplesList);
    indicationRequest->setRemoveTriples(removedIndicationTriplesList);
    */
}

void cutesib::Subscription::addTripleListInSubmodel(QList<librdf_statement *> tripleList)
{
    foreach (librdf_statement* statement, tripleList) {
        librdf_model_add_statement(submodel_, statement);
        librdf_free_statement(statement);
    }
}

void cutesib::Subscription::addTripleListInSubmodel(QList<cutesib::Triple *> tripleList)
{

}

