/**
* @author Ivan V. Galov <galov@cs.karelia.ru>
* @date   18.04.2016
* @brief  Handler for processing subscription operation.
*
* This file is part of the CuteSIB project.
*
* @section LICENSE
*
* CuteSIB is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* CuteSIB is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with CuteSIB; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor,
* Boston, MA  02110-1301  USA
*
* @section DESCRIPTION
*
* Handles subscription operation of SSAP protocol.
*
* Web-site: http://oss.fruct.org/wiki/SmartSpace-platform
* Copyright (C) 2015-2016 by PetrSU.
**/

#include "subscriptionhandler.h"

#include "scheduler.h"

#define wildcard1 	"sib:any"
#define wildcard2 	"http://www.nokia.com/NRC/M3/sib#any"

cutesib::SubscriptionHandler::SubscriptionHandler(QObject *parent) : OperationHandler(SUBSCRIPTIONHANDLER_ID, true, parent)
{

}

cutesib::SibError cutesib::SubscriptionHandler::initialize(cutesib::Context *context)
{
    SibModule::initialize(context);
    triplestore_ = this->context()->triplestore();
    RDF_world = triplestore_->getRDFWorld();
    RDF_model = triplestore_->getRDFModel();

    return SibError::NONE;
}

cutesib::CommandResponse *cutesib::SubscriptionHandler::subscribe(cutesib::CommandRequest *command)
{
    QList<Triple *> queryTriples;

    foreach (Triple* triple, command->queryTriples()) {
        queryTriples.append(triple->createCopy());
    }

    //Rdf Redland Initialize..
    librdf_statement* 	statement ;
    librdf_statement*	partial_statement;
    librdf_stream* stream;

    librdf_node* sub_node ;
    librdf_node* pred_node;
    librdf_node* obj_node ;

    QList<Triple *> tripleList;
    QList<librdf_statement*> statementsList;
    Triple* resultTriple;

    foreach (Triple* triple, queryTriples) {

        // Redland query with wildcard support
        partial_statement=librdf_new_statement(RDF_world);


        if ( (strcmp((char*) triple->subject().toUtf8().data(), wildcard1) ==0) || ( (strcmp((char*) triple->subject().toUtf8().data() ,wildcard2) ==0) ))
        {	librdf_statement_set_subject(partial_statement, NULL);
        }
        else
        {	librdf_statement_set_subject(partial_statement, librdf_new_node_from_uri_string(RDF_world,  (const unsigned char*) triple->subject().toUtf8().data()));
        }

        if ( (strcmp((char*) triple->predicate().toUtf8().data(), wildcard1) ==0) || ( (strcmp((char*) triple->predicate().toUtf8().data() ,wildcard2) ==0) ))
        {	librdf_statement_set_predicate(partial_statement, NULL);
        }
        else
        {	librdf_statement_set_predicate(partial_statement, librdf_new_node_from_uri_string(RDF_world, (const unsigned char*) triple->predicate().toUtf8().data()));
        }

        if ( (strcmp((char*) triple->object().toUtf8().data(), wildcard1) ==0) || ( (strcmp((char*) triple->object().toUtf8().data() ,wildcard2) ==0) ))
        {	librdf_statement_set_object(partial_statement, NULL);
        }
        else
        {
            if (triple->objectType() == RdfTripleType::URI)
            {	librdf_statement_set_object(partial_statement,
                                            librdf_new_node_from_uri_string(RDF_world, (const unsigned char*) triple->object().toUtf8().data()));
            }
            else
            {	librdf_statement_set_object(partial_statement,
                                            librdf_new_node_from_literal(RDF_world, (const unsigned char*) triple->object().toUtf8().data(), NULL, 0));
            }
        }

        /*if (p->virtuoso)
        stream=librdf_model_find_statements_in_context(p->RDF_model, partial_statement, p->virtuoso_context);
    else*/

        stream=librdf_model_find_statements(RDF_model, partial_statement);

        librdf_free_statement(partial_statement);


        //fprintf(stderr, "Query print all\n");
        while(!librdf_stream_end(stream)) {
            statement=librdf_stream_get_object(stream);


            ///////////////////////////////////////////////////////////////////////////////////////////

            sub_node  = librdf_statement_get_subject(statement);
            pred_node = librdf_statement_get_predicate(statement);
            obj_node  = librdf_statement_get_object(statement);

            if (( librdf_node_is_blank(sub_node))||( librdf_node_is_blank(obj_node)))
            {
                librdf_stream_next(stream);
                continue;
            }


            char* sub = (char*) librdf_uri_as_string( librdf_node_get_uri(sub_node));
            char* pred = (char*) librdf_uri_as_string( librdf_node_get_uri(pred_node));
            char* obj;

            if (librdf_node_is_literal(obj_node))
            {
                obj = (char*) librdf_node_get_literal_value(obj_node);
                //triple->objType = ssElement_TYPE_LIT;
                resultTriple = new Triple(QString(sub), QString(pred), QString(obj), RdfTripleType::LITERAL);
            }
            else
            {
                obj = (char*) librdf_uri_as_string(librdf_node_get_uri(obj_node));
                //triple->objType = ssElement_TYPE_URI;
                resultTriple = new Triple(QString(sub), QString(pred), QString(obj), RdfTripleType::URI);
            }

            tripleList.append(resultTriple);
            statementsList.append(librdf_new_statement_from_statement(statement));

            librdf_stream_next(stream);
        }

        librdf_free_stream(stream);

    }


    Subscription* subscription = new Subscription(command->subscriptionId(), queryTriples);
    subscription->addTripleListInSubmodel(statementsList);
    subscriptionsList.append(subscription);


    if (command->needConfirmation()) {
        CommandResponse* response = new CommandResponse(command, true);
        response->setTriples(tripleList);
        return response;
    }
    else {
        foreach (Triple* triple, tripleList) {
            delete triple;
        }
        return NULL;
    }
}

cutesib::CommandResponse *cutesib::SubscriptionHandler::unsubscribe(cutesib::CommandRequest *command)
{
    QString subscriptionId = command->subscriptionId();

    for (int i = 0; i < subscriptionsList.size(); i++) {
        if (subscriptionsList.at(i)->getSubscriptionId() == subscriptionId) {
            delete subscriptionsList.at(i);
            subscriptionsList.removeAt(i);
            break;
        }
    }

    if (command->needConfirmation())
        return new CommandResponse(command, true);
    else {
        delete command;
        return NULL;
    }
}

void cutesib::SubscriptionHandler::createSubscription(QString subscriptionId, QList<Triple*> queryList, QList<Triple*> initialTriples)
{

}

cutesib::CommandResponse *cutesib::SubscriptionHandler::perform(CommandRequest *command)
{
    CommandType type = command->Type();

    if (type == CommandType::SUBSCRIBE) {

        return subscribe(command);

    } else if (type == CommandType::UNSUBSCRIBE) {

        return unsubscribe(command);
    }

    return NULL;
}

void cutesib::SubscriptionHandler::updateState(cutesib::CommandRequest *command)
{
    QList<Triple*> insertedTriples = command->insertTriples();
    QList<Triple*> removedTriples = command->removeTriples();


    foreach (Subscription* subscription, subscriptionsList) {
        SubscriptionIndication indication;
        subscription->checkSubscription(insertedTriples, removedTriples, indication);

        /*
        qDebug("INSERTED:");
        foreach (Triple* triple, indication.insertedTriples) {
            printf("%s %s %s\n", triple->subject().toUtf8().data(),
                   triple->predicate().toUtf8().data(),
                   triple->object().toUtf8().data());
        }

        qDebug("REMOVED:");
        foreach (Triple* triple, indication.removedTriples) {
            printf("%s %s %s\n", triple->subject().toUtf8().data(),
                   triple->predicate().toUtf8().data(),
                   triple->object().toUtf8().data());
        }
        */

        if((!indication.insertedTriples.empty()) || (!indication.removedTriples.empty())) {
            // INFO: IVAN - save 'to' ID from initial subscription operation and set it here.
            CommandRequest* indicationRequest = new CommandRequest(CommandType::SUBSCRIPTION_INDICATE,
                                                                   this->id(), QString("PH-SSAP-XML"));

            indicationRequest->setSubscriptionId(subscription->getSubscriptionId());

            if (!indication.insertedTriples.empty())
                indicationRequest->setInsertTriples(indication.insertedTriples);

            if (!indication.removedTriples.empty())
                indicationRequest->setRemoveTriples(indication.removedTriples);

            context()->scheduler()->perform(indicationRequest);
        }
    }
}
