/**
 * @author Aleksandr A. Lomov <lomov@cs.karelia.ru>
 * @date   06.05.2015
 * @brief  Represents a RDF-triple.
 *
 * This file is part of the CuteSIB project.
 *
 * @section LICENSE
 *
 * CuteSIB is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * CuteSIB is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CuteSIB; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301  USA
 *
 * @section DESCRIPTION
 *
 * RDF triple consists of three elements: subject, predicate and object.
 * Predicate is also named property.
 * You can read more about RDF triples here: https://www.w3.org/TR/rdf11-primer/
 *
 * Now in the SIB the blank nodes (of RDF-graph) are not used that's why subject
 * is always URI and object can be URI or literal.
 *
 * Web-site: http://oss.fruct.org/wiki/SmartSpace-platform
 * Copyright (C) 2015-2016 by PetrSU.
 */


#ifndef CUTESIB_TRIPLE_H
#define CUTESIB_TRIPLE_H

#include <QString>


namespace cutesib {

/**
 * @brief Enum of RDF types for triple elements.
 * Now in the SIB the blank nodes (of RDF-graph) are not used that's why subject
 * is always URI and object can be URI or literal.
 */
enum RdfTripleType {
    UNKNOWN = -1,   ///< Unknown (error) type of element.
    URI,            ///< Element represents an URI (subject, predicate and object).
    LITERAL         ///< Element represents an literal (object).
};


/**
 * @brief The Triple class represents RDF-triple.
 * RDF triple consists of three elements: subject, predicate and object.
 * Predicate is also named property.
 * You can read more about RDF triples here: https://www.w3.org/TR/rdf11-primer/
 *
 * In the SIB triples-templates are used to query data.
 * Triple-template is a regular triple but for elements can be used special URI (#AnyUri).
 * This URI represents a mask (* - any value).
 */
class Triple
{
public:     // Fields.
    /**
     * @brief Any-URI represents a mask (* - any value) for triple-template in SIB.
     * Triple-template is used to query data in the SIB.
     *
     * Any-URI: http://www.nokia.com/NRC/M3/sib#any
     */
    static const QString AnyUri;

    /**
     * @brief Any-URI represents a mask (* - any value) for triple-template in SIB.
     * This is a short form of #AnyUri. It is used in the SIB only.
     */
    static const QString AnyShortForm;


public:     // Methods.
    /**
     * @brief Triple class constructor.
     * @param subject of the triple.
     * @param predicate of the triple.
     * @param object of the triple.
     * @param objecType is a type of object (can be URI or literal).
     * @see RdfTripleType
     */
    Triple(QString subject, QString predicate, QString object, RdfTripleType objecType);

    /**
     * @brief Triple class destructor.
     */
    ~Triple();

    /**
     * @brief Creates a new copy of the triple.
     * @return new triple with the same elements values and types of current triple.
     */
    Triple* createCopy();

    /**
     * @brief Gets the subject element of triple.
     * @return subject value.
     */
    QString subject() const { return subject_; }

    /**
     * @brief Gets the predicate element of triple.
     * @return predicate value.
     */
    QString predicate() const { return predicate_; }

    /**
     * @brief Gets the object element of triple.
     * @return object value.
     */
    QString object() const {return object_; }

    /**
     * @brief Gets the type of triple object.
     * @return type of object.
     */
    RdfTripleType objectType() const { return objectType_; }


private:    // Fields.
    QString subject_;       ///< Subject element of the triple.
    QString predicate_;     ///< Predicate (property) element of the triple.
    QString object_;        ///< Object element of the triple.
    RdfTripleType objectType_;  ///< Type of the object element.

    // NOTE: Debug - remove in stable.
    static int instanceCounter;
};

} // Namespace cutesib.


#endif // CUTESIB_TRIPLE_H
