/**
 * @author Aleksandr A. Lomov <lomov@cs.karelia.ru>
 * @date   24.05.2016
 * @brief  Loader for discovery access point.
 *
 * This file is part of the CuteSIB project.
 *
 * @section LICENSE
 *
 * CuteSIB is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * CuteSIB is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CuteSIB; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301  USA
 *
 * @section DESCRIPTION
 *
 * Loader for tcp access points.
 * Loader is a factory to create access point instances.
 * Also loader contains information about access points (version, protocol etc.).
 *
 * Web-site: http://oss.fruct.org/wiki/SmartSpace-platform
 * Copyright (C) 2015-2016 by PetrSU.
 */


#ifndef CUTESIB_TCPACCESSPOINTLOADER_H
#define CUTESIB_TCPACCESSPOINTLOADER_H

#include "accesspointloaderinterface.h"
#include "tcpaccesspoint.h"


namespace cutesib {
namespace access_point {

/**
 * @brief The TcpAccessPointLoader class is factory to create tcp access point instances.
 * @see cutesib::access_point::AccessPointLoaderInterface
 */
class TcpAccessPointLoader : public QObject, public AccessPointLoaderInterface
{
    Q_OBJECT
    Q_INTERFACES(cutesib::access_point::AccessPointLoaderInterface)

#if QT_VERSION >= 0x050000
    Q_PLUGIN_METADATA(IID TcpAccessPoint_iid FILE CUTESIB_TCPAP_SSAPXML_PLUGIN_FILE)
#endif

public:
    /**
     * @brief TcpAccessPointLoader constructor.
     */
    TcpAccessPointLoader();

    /**
     * @brief Gets identifier of loader.
     * @return identifier of loader.
     */
    QString id();

    /**
     * @brief Gets a description of access points that are created by loader.
     * @return description.
     */
    QString description();

    /**
     * @brief Gets a protocols that are supported by access points.
     * Access point supports SSAP/XML protocol.
     * @return string with supported protocols.
     */
    QString protocols();

    /**
     * @brief Gets a version of loader and access points.
     * @return version.
     */
    QString version();

    /**
     * @brief Creates an access point.
     * Format of parameters is key-value pairs: 'key:value,key:value,...'.
     * There is one parameter: port (default value is 10010).
     * @param accessPointId is an identifier of new access point.
     * @param parameters for the access point.
     * @return created accsss point on success or nullptr otherwise.
     */
    AccessPointInterface* create(QString accessPointId, QString parameters);

private:
    static const QString id_;               ///< Identifier of loader (plug-in).
    static const QString version_;          ///< Version of loader and access points.
    static const QString description_;      ///< Description of access points.
    static const QString protocols_;        ///< Protocols that are supported by access points.
};

} // namespace access_point
} // namespace cutesib


#endif // CUTESIB_TCPACCESSPOINTLOADER_H
