/**
 * @author Aleksandr A. Lomov <lomov@cs.karelia.ru>
 * @date   05.03.2015
 * @brief  Class for TCP client.
 *
 * This file is part of the CuteSIB project.
 *
 * @section LICENSE
 *
 * CuteSIB is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * CuteSIB is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CuteSIB; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301  USA
 *
 * @section DESCRIPTION
 *
 * Class for TCP client.
 *
 * Contact: http://oss.fruct.org/wiki/SmartSpace-platform
 * Copyright (C) 2015 by PetrSU.
 */


#ifndef TCPCLIENT_H
#define TCPCLIENT_H

#include "tcpaccesspoint_global.h"

#include <QtCore/QObject>
#include <QtCore/QBuffer>
#include <QtCore/QMutex>
#include <QtNetwork/QTcpSocket>

#include "client.h"
#include "tcpaccesspoint.h"

namespace cutesib {
namespace access_point {

// Forward declaration.
class TcpAccessPoint;

/**
 * @brief The TcpClient class represents a client for TCP access point.
 * @see Client
 */
class TcpClient : public Client
{
    Q_OBJECT

public: // Fields.
    static const QByteArray MessageEnding; ///< Ending of SSAP/XML message (</SSAP_message>).


public: // Methods.
    /**
     * @brief TcpClient class constructor.
     * @param accessPoint of client.
     * @param socket of client.
     */
    TcpClient(TcpAccessPoint *accessPoint, QTcpSocket* socket);

    /**
     * @brief TcpClient class destructor.
     */
    ~TcpClient();

    /**
     * @brief Gets socket of client.
     * @return socket.
     */
    QTcpSocket* getSocket() const;

    /**
     * @brief Gets access point of client.
     * @return access point.
     * @see AccessPointInterface
     */
    AccessPointInterface *getAccessPoint();

    /**
     * @brief Checks: is buffer for received message has a full message.
     * Full message means message with start (<SSAP_message>) and end tag (<SSAP_message>).
     * @return true if the message is full or false otherwise.
     */
    bool isReceivedFullMessage();

    /**
     * @brief Adds to buffer new data.
     * @param data to add to buffer for received messages.
     */
    void addMessageData(QByteArray data);

    /**
     * @brief Gets a message frrom buffer.
     * Gets a first full message from buffer and
     * removes it from buffer.
     * @return data represent the full message or empty array if the message is not full.
     */
    QByteArray getMessage();

    /**
     * @brief Sends data to this client.
     * @param message to send.
     * @return 0 on success or not otherwise.
     */
    int send(QByteArray message);

    /**
     * @brief Sets flag about connection state.
     * @param isConnected flag.
     */
    void isConnected(bool isConnected) { isConnected_ = isConnected; }

    /**
     * @brief Gets flag about connection state.
     * @return true if the client is connect or false otherwise.
     */
    bool isConnected() { return isConnected_; }

    void close() /*override*/;


    ///TODO: check methods.
    void Lock();
    void Unlock();

    ///TODO: check (use with signals).
    void setMessage(QByteArray data);
    QByteArray getMessagePart(qint64 size);


private:
    QTcpSocket* socket_;        ///< Socket of the client.
    QMutex socketLock_;         ///< Lock for asynchronous usage (now is not used).
    TcpAccessPoint *accessPoint_; ///< Access point of the client.

    QByteArray receiveBuffer_;  ///< Buffer for received messages.
    QBuffer messageBuffer_;     ///< Buffer for sending message.

    bool isConnected_;          ///< Flag is client connected or not.
};

} // namespace access_point
} // namespace cutesib

#endif // TCPCLIENT_H
