/**
* @author Ivan V. Galov <galov@cs.karelia.ru>
* @date   18.04.2016
* @brief  Class for creating and working with Redland triplestore.
*
* This file is part of the CuteSIB project.
*
* @section LICENSE
*
* CuteSIB is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* CuteSIB is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with CuteSIB; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor,
* Boston, MA  02110-1301  USA
*
* @section DESCRIPTION
*
* Class for creating and working with Redland triplestore.
*
* Web-site: http://oss.fruct.org/wiki/SmartSpace-platform
* Copyright (C) 2015-2016 by PetrSU.
**/

#include "triplestore.h"
#include <QFile>
#include <QDir>

Triplestore::Triplestore(cutesib::SibConfig *config)
{
    config_ = config;

    // storage types: memory, hash-memory, hash-bdb, sqlite, virtuoso
    ssType_ = config->getTriplestoreParameter("type");

    if (ssType_.isNull() || ssType_.isEmpty()) {
        qWarning("Storage type isn't defined. Default is used.");
        ssType_ = "hash-bdb";
    }

    ssName_ = config->getTriplestoreParameter("name");

    if (ssName_.isNull()) {
        qWarning("Storage name isn't defined. Default is used.");
        ssName_ = "X";

    }

    ssPath_ = config->getTriplestoreParameter("path");

    if (ssPath_.isNull() || ssPath_.isEmpty()) {
        qWarning("Storage path isn't defined. Current directory is used.");
        ssPath_ = ".";
    }
    // TODO: check for MS Windows path
    if (!ssPath_.endsWith("/")) {
        ssPath_.append("/");
    }

    if (!QDir(ssPath_).exists()) {
        qFatal("Path for storage does not exists!");
    }

    dbHost_ = config->getTriplestoreParameter("host");
    dbUser_ = config->getTriplestoreParameter("user");
    dbPassword_ = config->getTriplestoreParameter("password");
}

Triplestore::~Triplestore()
{
    librdf_free_model(RDF_model);
    librdf_free_storage(RDF_storage);
}

void Triplestore::initialize()
{
    //INITIALIZE REDLAND ENVIROMENT
    RDF_world = librdf_new_world();
    librdf_world_open(RDF_world);
// TODO: add ifdef (old libs of redland)
    //rasqal_world_set_warning_level(librdf_world_get_rasqal(RDF_world), 0);


    //////Creating Model and storage ////////////////////////
    ////////////////////////////////////////////////////////

    if (ssType_ == "memory") {
        RDF_storage = librdf_new_storage(RDF_world,  NULL, NULL, NULL);
    }
    else if (ssType_ == "hash-memory") {
        RDF_storage = librdf_new_storage(RDF_world, "hashes", NULL, "hash-type='memory'");
    }
    else if (ssType_ == "hash-bdb") {
        QString sp2o = ssPath_ + ssName_ + QString("-sp2o.db");
        QString po2s = ssPath_ + ssName_ + QString("-so2p.db");
        QString so2p = ssPath_ + ssName_ + QString("-po2s.db");
        QString newStorageFlag = "yes";
        if (
                (
                (QFile(sp2o).exists()) &&
                (QFile(so2p).exists()) &&
                (QFile(po2s).exists())
                ) /*&&
                (enable_rdf_pp == FALSE)*/
                )
        {
            qDebug("Found a BDB local file, re-charging.. \n");
            //RDF_storage = librdf_new_storage(RDF_world, "hashes", ssName_.toUtf8().data(), "new='no',hash-type='bdb',dir='.'");
            newStorageFlag = "no";
        }
        else
        {
            //RDF_storage = librdf_new_storage(RDF_world, "hashes", ssName_.toUtf8().data(), "new='yes',hash-type='bdb',dir='.'");
            newStorageFlag = "yes";
        }
        QString paramString = QString("new='%1',hash-type='bdb',dir='%2'").arg(newStorageFlag).arg(ssPath_);
        RDF_storage = librdf_new_storage(RDF_world, "hashes", ssName_.toUtf8().data(), paramString.toUtf8().data());

    }
    /*
    else if (ssType_ == "file") {
        RDF_storage = librdf_new_storage(RDF_world, "file", "thing.rdf", NULL);
    }
    */
    else if (ssType_ == "sqlite") {
        if (QFile(ssName_).exists()) {
            RDF_storage = librdf_new_storage(RDF_world, "sqlite", ssName_.toUtf8().data(), NULL);
        }
        else {
            RDF_storage = librdf_new_storage(RDF_world, "sqlite", ssName_.toUtf8().data(), "new='yes'");
        }
    }
    /*
    else if (ssType_ == "mysql") {

    }
    */
    else if (ssType_ == "virtuoso") {
        QString  uriDbName = QString("http://sib#%1").arg(ssName_);

        //librdf_hash_put_strings(options, "contexts", "yes");
        //virtuosoContext = librdf_new_node_from_uri_string(RDF_world, (const unsigned char *) uriDbName.toUtf8().data());

        QString params = QString("dsn='VOS', host='%1', user='%2', password='%3'").arg(dbHost_).arg(dbUser_).arg(dbPassword_);
        RDF_storage = librdf_new_storage(RDF_world, "virtuoso", uriDbName.toUtf8().data(),
                                                      params.toUtf8().data());
    }

    RDF_model = librdf_new_model(RDF_world, RDF_storage , NULL);
}

void Triplestore::resetStorageModel()
{
    /*
    //MAIN TRIPLESTORE
    if (sd->mem_volatile)
    {
        // Cleaning ////////////////////////
        librdf_free_model(sd->RDF_model);
        librdf_free_storage(sd->RDF_storage);
        ////////////////////////////////////

        sd->RDF_storage=librdf_new_storage(sd->RDF_world,  NULL, NULL, NULL);
        sd->RDF_model=librdf_new_model(sd->RDF_world, sd->RDF_storage , NULL);
    }
    else if (sd->mem_volatile_hash)
    {
        // Cleaning ////////////////////////
        librdf_free_model(sd->RDF_model);
        librdf_free_storage(sd->RDF_storage);
        ////////////////////////////////////

        sd->RDF_storage=librdf_new_storage(sd->RDF_world, "hashes", NULL, "hash-type='memory'");
        sd->RDF_model=librdf_new_model(sd->RDF_world, sd->RDF_storage , NULL);
    }
    else
    {
        if (sd->sqlite)
        {
            // Cleaning ////////////////////////
            librdf_free_model(sd->RDF_model);
            librdf_free_storage(sd->RDF_storage);
            ////////////////////////////////////
            sd->RDF_storage=librdf_new_storage(sd->RDF_world, "sqlite", sd->ss_name, "new='yes'");
            sd->RDF_model=librdf_new_model(sd->RDF_world, sd->RDF_storage , NULL);
        }
        else
        {
            if (sd->virtuoso)
            {
                librdf_model_context_remove_statements(sd->RDF_model, sd->virtuoso_context);
            }
            else
            {
*/
                //BDB default

                // Cleaning ////////////////////////

                //librdf_free_model(RDF_model);
                //librdf_free_storage(RDF_storage);


//                librdf_free_world(RDF_world);

//                RDF_world=librdf_new_world();
//                librdf_world_open(RDF_world);
//                rasqal_world_set_warning_level(librdf_world_get_rasqal(RDF_world), 0);


                //RDF_storage = librdf_new_storage(RDF_world, "hashes", ssName_.toUtf8().data(), "new='yes',hash-type='bdb',dir='.'");
                //RDF_model = librdf_new_model(RDF_world, RDF_storage , NULL);
                //initialize();
/*            }
        }
    }*/
}

void Triplestore::removeAllStatements()
{
    librdf_statement* statement;
    librdf_stream *stream = librdf_model_as_stream(RDF_model);
    if (stream != NULL) {
        while(!librdf_stream_end(stream))
        {
            statement = librdf_stream_get_object(stream);

            librdf_model_remove_statement(RDF_model, statement);

            librdf_stream_next(stream);
        }

        librdf_free_stream(stream);
    }
}

librdf_world *Triplestore::getRDFWorld()
{
    return RDF_world;
}

librdf_model *Triplestore::getRDFModel()
{
    return RDF_model;
}

cutesib::SibConfig *Triplestore::getConfig()
{
    return config_;
}
